from __future__ import annotations

"""
App Oauth Client
"""
import json
import contextlib
from abc import ABC
from urllib.parse import urljoin
from lazyops.libs import lazyload
from fastapi import Request
from fastapi.background import BackgroundTasks
from ..utils.lazy import get_az_settings, get_az_mtg_api, get_az_resource_schema, logger
from ..utils.helpers import get_hashed_key, create_code_challenge, parse_scopes, encode_params_to_url


from typing import Optional, List, Dict, Any, Union, Type

if lazyload.TYPE_CHECKING:
    import niquests
    from fastapi import FastAPI
    from fastapi.background import BackgroundTasks
    from starlette.datastructures import URLPath
    from ..configs import AuthZeroSettings
    from ..flows.admin import AZManagementAPI
    from ..types.clients import AuthZeroClientObject
    from ..types.current_user import CurrentUser
    
else:
    niquests = lazyload.LazyLoad("niquests")


class AuthZeroOAuthClient(ABC):
    """
    The Auth Zero OAuth Client
    """

    _settings: Optional['AuthZeroSettings'] = None
    _mtg_api: Optional['AZManagementAPI'] = None

    def __init__(
        self,
        app: 'FastAPI',
        client_id: Optional[str] = None,
        secret_key: Optional[str] = None,
        app_ingress: Optional[str] = None,
        callback_path: Optional[str] = '/auth/callback',
        logout_path: Optional[str] = None,
        include_wildcard_origin: Optional[bool] = True,
        user_class: Optional[Type['CurrentUser']] = None,
        **kwargs,
    ):
        """
        Initializes the Auth Zero OAuth Client
        """
        self.app = app
        self.client_id = client_id or self.settings.app_client_id or self.settings.client_id
        if app_ingress: self.settings.configure(app_ingress = app_ingress)
        self.secret_key = secret_key or get_hashed_key(self.settings.app_domain + self.settings.client_id)[:64]
        self.code_challenge = create_code_challenge(self.secret_key)
        self.callback_path = callback_path
        self.logout_path = logout_path
        self.include_wildcard_origin = include_wildcard_origin
        self.redirect_uri = urljoin(self.settings.app_ingress, self.callback_path)
        self.user_class = user_class or get_az_resource_schema('current_user')

        # Extra Properties
        self.docs_schema_index: Dict[str, str] = {}
        self.source_openapi_schema: Dict[str, Any] = None


    """
    Primary Methods
    """

    async def authorize_app_in_authzero(
        self,
        ingress_url: Optional[str] = None,
        include_wildcard_origin: Optional[bool] = None,
        verbose: Optional[bool] = True,
        **kwargs,
    ) -> 'AuthZeroClientObject':  # sourcery skip: low-code-quality
        """
        Authorizes the app in Auth0

        - Ensures that the app's ingress, callback path, and logout path are configured
        """
        az_client = await self.mtg_api.aget_az_client(self.client_id)
        source_counts = az_client.get_app_update_counts()

        if ingress_url is None: ingress_url = self.settings.app_ingress
        if include_wildcard_origin is None: include_wildcard_origin = self.include_wildcard_origin
        if include_wildcard_origin: include_wildcard_origin = ('localhost' not in ingress_url and '0.0.0' not in ingress_url)
        if include_wildcard_origin:
            wildcard_origin = ingress_url.replace('://', '://*.', 1)
            az_client.add_app_url(allowed_origin=wildcard_origin, verbose=verbose)

        elif 'localhost' not in ingress_url:
            az_client.add_app_url(allowed_origin=ingress_url, verbose=verbose)
        
        callback_url = (
            self.callback_path
            if self.callback_path.startswith('http')
            else urljoin(ingress_url, self.callback_path)
        )
        
        logout_url = (
            self.logout_path
            if self.logout_path.startswith('http')
            else urljoin(ingress_url, self.logout_path)
        ) if self.logout_path else ingress_url
        
        az_client.add_app_url(callback = callback_url, allowed_logout_url = logout_url, web_origin = ingress_url, verbose = verbose)
        
        if not az_client._needs_update:
            if verbose: logger.info(f'Client: `{az_client.name}` is already authorized with `{ingress_url}`', colored = True, prefix = f'|g|{az_client.client_id}|e|')
            return az_client

        response = await self.mtg_api.ahpatch(f'clients/{az_client.client_id}', json = az_client.get_app_patch_data())
        if response.status_code > 299:
            logger.warning(f'[{response.status_code}] Error updating app: {response.text} for {az_client.client_id}')
            return az_client
        
        resp = response.json()
        if verbose: 
            logger.info(f'Updated Client: `{az_client.name}`\n\n{resp}', colored = True, prefix = f'|g|{az_client.client_id}|e|')
            new_counts = az_client.get_app_update_counts()
            for key, value in new_counts.items():
                if value > source_counts[key]:
                    logger.info(f'{key}: {value} -> {source_counts[key]} (+{value - source_counts[key]})', colored = True, prefix = f'|g|{az_client.client_id}|e|')
    
        return az_client
            
    
    def get_authorization_redirect_url(
        self,
        scope: Optional[str] = None,
        scopes: Optional[List[str]] = None,
        audience: Optional[str] = None,
        **kwargs
    ) -> str:
        """
        Returns the Authorization Redirect URL
        """
        scopes = parse_scopes(scope = scope, scopes = scopes)
        assert scopes, 'At least one scope must be provided'
        scope = ' '.join(scopes)
        if audience is None: audience = self.settings.audience
        assert audience, 'Audience must be provided'
        params = {
            'response_type': 'code',
            'code_challenge': self.code_challenge,
            'code_challenge_method': 'S256',
            'client_id': self.client_id,
            'redirect_uri': self.redirect_uri,
            'scope': scope,
            'audience': audience
        }
        if kwargs: 
            for k, v in kwargs.items():
                if k in params and v: params[k] = v
        return encode_params_to_url(params, self.authorize_url)
    

    async def authorize_app_user(
        self,
        request: 'Request',
        code: str,
        scope: Optional[str] = None,
        scopes: Optional[List[str]] = None,
        background_tasks: Optional['BackgroundTasks'] = None,
        user_class: Optional[Type[CurrentUser]] = None,
        **kwargs,
    ) -> 'CurrentUser':
        """
        Authorizes the App User
        """
        data = {
            'grant_type': 'authorization_code',
            'client_id': self.client_id,
            'code_verifier': self.secret_key,
            'code': code,
            'redirect_uri': self.redirect_uri
        }
        response = await self.mtg_api.ahpost(self.settings.oauth_url, data = data)
        try:
            response.raise_for_status()
        except Exception as e:
            logger.error(f'Error Authorizing User: {e}')
            from ..types.errors import AuthorizationException
            raise AuthorizationException(detail = response.text, status_code = response.status_code) from e
        
        user_data = response.json()
        if user_class is None: user_class = self.user_class
        current_user = user_class()
        try:
            await current_user.from_access_token(
                access_token = user_data['access_token'],
                request = request,
                scope = scope,
                scopes = scopes,
                background_tasks = background_tasks,
                **kwargs,
            )
        except Exception as e:
            logger.error(f'Error Authorizing User: {e}')
            raise e
        
        return current_user
    
    """
    Helper Methods
    """

    def get_app_redirection(
        self,
        redirect: str
    ) -> Union[str, 'URLPath']:
        """
        Gets the app redirection
        """
        if redirect.startswith('http'): return redirect
        if 'docs=' in redirect:
            base_url = str(self.app.url_path_for('docs').make_absolute_url(self.app_ingress)) + '#/operations'
            redirect = redirect.replace('docs=', '')
            if redirect in self.docs_schema_index:
                return f'{base_url}/{self.docs_schema_index[redirect]}'
        return self.app.url_path_for(redirect).make_absolute_url(self.app_ingress)


    def create_docs_index(self, schema: Dict[str, Any]):
        """
        Creates the docs index
        """
        for path in schema.get('paths', {}):
            for method in schema['paths'][path]:
                if 'operationId' in schema['paths'][path][method]:
                    doc_name = schema['paths'][path][method]['summary'].replace(' ', '').lower()
                    self.docs_schema_index[doc_name] = schema['paths'][path][method]['operationId']


    def create_openapi_source_spec(self, spec: Dict[str, Any], spec_map: Optional[Dict[str, Any]] = None):
        """
        Creates the Source Spec

        - Handles some custom logic for the OpenAPI Spec
            Namely:
            - AppResponse-Input -> AppResponse
            - AppResponse-Output -> AppResponse
        """
        if not spec_map: return
        _spec = json.dumps(spec)
        for key, value in spec_map.items():
            _spec = _spec.replace(key, value)
        self.source_openapi_schema = json.loads(_spec)


    def mount_oauth_components(
        self, 
        login_path: Optional[str] = '/login',
        logout_path: Optional[str] = '/logout',

        server_identity: Optional[str] = None,
        server_identity_path: Optional[str] = '/_identity',

        enable_authorize: Optional[bool] = True,
        authorize_path: Optional[str] = '/authorize',

        enable_whoami: Optional[bool] = None,

        include_in_schema: Optional[bool] = None,
        user_class: Optional[Type['CurrentUser']] = None,
    ):
        """
        Mounts the OAuth Components
        """
        from fastapi import APIRouter, Depends, Query
        from fastapi.responses import RedirectResponse, JSONResponse, PlainTextResponse
        from .dependencies import get_current_user, CurrentUser as _CurrentUser

        router = APIRouter()

        @router.get(login_path, include_in_schema = include_in_schema)
        async def login(
            current_user: Optional[CurrentUser] = Depends(get_current_user(required = False, user_class = user_class)),
            redirect: Optional[str] = Query(None, description="The redirect page to use after login"),
        ):
            """
            Login Endpoint
            """
            if current_user is not None and current_user.is_valid:
                if redirect is not None: return RedirectResponse(self.get_app_redirection(redirect))
                return {'login': 'success', 'x-api-key': current_user.api_key}
            
            redirect_url = self.get_authorization_redirect_url(
                audience = self.settings.audience,
                scopes = self.settings.app_scopes,
            )
            response = RedirectResponse(redirect_url)
            if redirect: response.set_cookie(
                key = 'x-auth-redirect', value = redirect, max_age = 60, httponly = True,
            )
            return response
        
        @router.get(self.callback_path, include_in_schema = include_in_schema)
        async def auth_callback(
            request: Request,
            code: Optional[str] = Query(None),
            background_tasks: BackgroundTasks = BackgroundTasks,
        ):
            """
            Callback Endpoint
            """
            current_user = await self.authorize_app_user(
                request = request,
                code = code,
                scopes = self.settings.app_scopes,
                background_tasks = background_tasks,
            )
            if self.settings.is_development_env:
                logger.info(f'User {current_user.user_id} logged in')

            if redir_value := request.cookies.get('x-auth-redirect'):
                redirect = self.get_app_redirection(redir_value)
                if self.settings.is_development_env:
                    logger.info(f'Found redirect cookie: {redir_value} - Redirecting to {redirect}')
                response = RedirectResponse(redirect)
                response.delete_cookie('x-auth-redirect')
            else:
                response = JSONResponse({'login': 'success', 'x-api-key': current_user.api_key})
            response.set_cookie(**current_user.get_session_cookie_kwargs())
            return response

        @router.get(logout_path, include_in_schema = include_in_schema)
        async def logout(
            current_user: Optional[_CurrentUser] = Depends(get_current_user(required = False, user_class = user_class)),
        ):
            """
            Logout Endpoint
            """
            if current_user is None: return {'logout': 'no_user_found'}
            response = JSONResponse({'logout': 'success'})
            response.delete_cookie(**current_user.get_session_cookie_kwargs(is_delete = True))
            return response
        
        @router.get(self.callback_path, include_in_schema = include_in_schema)
        async def get_api_key(
            current_user: Optional[_CurrentUser] = Depends(get_current_user(required = False, user_class = user_class)),
            plaintext: Optional[bool] = Query(None, description="If True, will return the api key in plaintext"),
        ):
            """
            Get the API Key
            """
            if current_user is None or not current_user.is_valid: 
                return 'null' if plaintext else {'api_key': 'no_user_found'}
            response = PlainTextResponse(content = current_user.api_key) if plaintext else JSONResponse(content = {'api_key': current_user.api_key})
            response.set_cookie(**current_user.get_session_cookie_kwargs())
            return response
        
        if server_identity:
            @router.get(server_identity_path, include_in_schema = include_in_schema)
            async def get_server_identity(
                request: Request,
            ):
                """
                Get the Server Identity
                """
                return PlainTextResponse(content = server_identity)
        
        if enable_authorize:
            @router.get(authorize_path, include_in_schema = include_in_schema)
            async def authorize_user(
                current_user: _CurrentUser = Depends(get_current_user(user_class = user_class)),
            ):
                """
                Authorize the User or Client API by configuring the Cookies
                """
                if not current_user.is_valid: return {'authorize': 'invalid_user'}
                response = JSONResponse(content = {'authorized': True, 'identity': server_identity or self.settings.app_name, 'environment': self.settings.app_env.value, 'api-key': current_user.api_key})
                response.set_cookie(**current_user.get_session_cookie_kwargs())
                return response

        if enable_whoami or self.settings.is_local_env:
            @router.get('/whoami')
            async def get_whoami_for_user(
                current_user: Optional[_CurrentUser] = Depends(get_current_user(required = False, user_class = user_class)),
                pretty: Optional[bool] = Query(None, description="If True, will return the user in a pretty format"),
            ):
                """
                Get the Whoami Data for the User
                """
                if current_user is None: return {'whoami': 'no_user_found'}
                data = current_user.get_whoami_data()
                if pretty: 
                    import yaml
                    data = yaml.dump(data, default_flow_style = False, indent = 2)
                    return PlainTextResponse(content = data)
                return data

        self.app.include_router(router, tags = ['oauth'])
        

        # @contextlib.asynccontextmanager
        # async def oauth_lifespan(app: 'APIRouter'):


        # @router.on_event("")


    """
    Properties
    """

    @property
    def settings(self) -> 'AuthZeroSettings':
        """
        Returns the settings
        """
        if self._settings is None: self._settings = get_az_settings()
        return self._settings

    @property
    def mtg_api(self) -> 'AZManagementAPI':
        """
        Returns the AZ Management API
        """
        if self._mtg_api is None: self._mtg_api = get_az_mtg_api()
        return self._mtg_api
    
    @property
    def app_ingress(self) -> str:
        """
        Returns the App Ingress
        """
        return self.settings.app_ingress
    
    @property
    def authorize_url(self) -> str:
        """
        Returns the Authorize URL
        """
        return self.settings.authorize_url



from __future__ import annotations

"""
Web Browser Api using Playwright
"""
import asyncio
# from lazyops.libs.proxyobj import ProxyObject, proxied
from typing import Optional, Dict, Any, List, Union, Type, TYPE_CHECKING
from .base import BaseGlobalClient

if TYPE_CHECKING:
    from playwright.sync_api import PlaywrightContextManager, Playwright, Browser as SyncBrowser
    from playwright.async_api import (
        PlaywrightContextManager as AsyncPlaywrightContextManager, 
        Playwright as AsyncPlaywright,
        Browser as AsyncBrowser
    )

# @proxied
class BrowserClient(BaseGlobalClient):
    """
    The Browser Client Meta Class
    """
    initialized: bool = False
    _initializing: bool = False

    _abrowser: 'AsyncBrowser' = None
    _apw_client: 'AsyncPlaywright' = None
    _apw_context_manager: 'AsyncPlaywrightContextManager' = None

    _browser: 'SyncBrowser' = None
    _pw_client: 'Playwright' = None
    _pw_context_manager: 'PlaywrightContextManager' = None
    
    @property
    def abrowser(self) -> 'AsyncBrowser':
        """
        Returns the browser instance
        """
        if self._abrowser is None:
            raise RuntimeError('AsyncBrowser not initialized. Initialize with Browser.ainit()')
        return self._abrowser
    
    @property
    def acontext_manager(self) -> 'AsyncPlaywrightContextManager':
        """
        Returns the Playwright context manager
        """
        if self._apw_context_manager is None:
            from playwright.async_api import async_playwright
            self._apw_context_manager = async_playwright()
        return self._apw_context_manager
    
    @property
    def aclient(self) -> 'AsyncPlaywright':
        """
        Returns the Playwright instance
        """
        if self._apw_client is None:
            raise RuntimeError('AsyncPlaywright not initialized. Initialize with Browser.ainit()')
        return self._apw_client

    @property
    def browser(self) -> 'SyncBrowser':
        """
        Returns the browser instance
        """
        if self._browser is None:
            self._browser = self.client.chromium.launch(
                headless = True,
                downloads_path = self.settings.module_path.joinpath('data').as_posix()
            )
        return self._browser
    
    @property
    def context_manager(self) -> 'PlaywrightContextManager':
        """
        Returns the Playwright context manager
        """
        if self._pw_context_manager is None:
            from playwright.sync_api import sync_playwright
            self._pw_context_manager = sync_playwright()
        return self._pw_context_manager
    
    @property
    def client(self) -> 'Playwright':
        """
        Returns the Playwright instance
        """
        if self._pw_client is None:
            self._pw_client = self.context_manager.start()
        return self._pw_client
    
    def _set_apw_client(self, task: asyncio.Task):
        """
        Sets the AsyncPlaywright instance
        """
        self.logger.info('Setting AsyncPlaywright')
        self._apw_client = task.result()

    def _set_abrowser(self, task: asyncio.Task):
        """
        Sets the AsyncBrowser instance
        """
        self.logger.info('Setting AsyncBrowser')
        self._abrowser = task.result()

    async def ainit(self):
        """
        Initializes the browser
        """
        if self.initialized: return
        self.logger.info('Initializing Browser')
        self._initializing = True
        if self._apw_client is None:
            self._apw_client = await self.acontext_manager.start()
        if self._abrowser is None:
            self._abrowser = await self.aclient.chromium.launch(
                headless = True,
                downloads_path = self.settings.module_path.joinpath('data').as_posix()
            )
        self.initialized = True
        self._initializing = False
    
    def close(self):
        """
        Closes the browser
        """
        if self._abrowser:
            self.pooler.create_background(
                self._abrowser.close,
            )
            self._abrowser = None
        if self._browser:
            self._browser.close()
            self._browser = None
        
        if self._apw_client:
            self.pooler.create_background(
                self._apw_client.stop,
            )
            self._apw_client = None
        if self._pw_client:
            self._pw_client.stop()
            self._pw_client = None
        
        self._apw_context_manager = None
        self._pw_context_manager = None
        self.initialized = False



    async def aclose(self):
        """
        Closes the browser
        """
        if self._abrowser:
            await self._abrowser.close()
            self._abrowser = None
        if self._browser:
            self._browser.close()
            self._browser = None
        
        if self._apw_client:
            await self._apw_client.stop()
            self._apw_client = None
        if self._pw_client:
            self._pw_client.stop()
            self._pw_client = None
        
        self._apw_context_manager = None
        self._pw_context_manager = None
        self.initialized = False


# Browser: BrowserClient = ProxyObject(
 #    obj_cls = BrowserClient,
# )




from __future__ import annotations

"""
Vector Types
"""

from enum import Enum
from typing import Optional, List, Dict, Any, Union, Literal, TYPE_CHECKING

class VectorDistance(str, Enum):
    """

    Euclidean (L2): `<->`
    Cosine: `<=>`
    Max Inner Product: `<~>` Fallback to Manhattan (L1): `<~>`
    Manhattan (L1): `<~>`
    """
    EUCLIDEAN = '<->'
    COSINE = '<=>'
    MAX_INNER_PRODUCT = '<#>'
    # INNER_DISTANCE = '<#>'
    # MAX_INNER_PRODUCT = '<~>'
    # INNER_DISTANCE = '<~>'
    # MANHATTAN = '<~>'

    @property
    def vector_op(self) -> str:
        """
        Returns the Vector Op Name

        vector_l2_ops = L2 distance
        vector_ip_ops = Inner Product
        vector_cosine_ops = Cosine Distance
        """
        if self == VectorDistance.EUCLIDEAN:
            return 'vector_l2_ops'
        elif self == VectorDistance.COSINE:
            return 'vector_cosine_ops'
        return 'vector_ip_ops'
    
    @classmethod
    def parse(cls, value: str) -> 'VectorDistance':
        """
        Parses the Vector Distance
        """
        value = value.lower()
        if value in {'euclidean', 'vector_l2_ops', '<->'}:
            return cls.EUCLIDEAN
        if value in {'cosine', 'vector_cosine_ops', '<=>'}:
            return cls.COSINE
        if value in {'max_inner_product', 'vector_ip_ops', '<#>', 'inner_product'}:
            return cls.MAX_INNER_PRODUCT
        # if value in {'inner_distance', '<~>'}:
        #     return cls.INNER_DISTANCE
        # if value == 'manhattan':
        #     return cls.MANHATTAN
        raise ValueError(f"Invalid/Unsupported Vector Distance: {value}")




"""
Import Handler for niquests
"""

from lazyops.utils.imports import resolve_missing, require_missing_wrapper

try:
    import niquests
    _niquests_available = True
except ImportError:
    niquests = object
    _niquests_available = False

def resolve_niquests(
    required: bool = False,
):
    """
    Ensures that `niquests` is available
    """
    global niquests, _niquests_available
    if not _niquests_available:
        resolve_missing('niquests', required = required)
        import niquests
        _niquests_available = True
        globals()['niquests'] = niquests


def require_niquests(
    required: bool = False,
):
    """
    Wrapper for `resolve_niquests` that can be used as a decorator
    """
    def decorator(func):
        return require_missing_wrapper(
            resolver = resolve_niquests, 
            func = func, 
            required = required
        )
    return decorator


from __future__ import annotations

"""
Admin / Management API
"""

import inspect
from abc import ABC
from urllib.parse import urljoin
from fastapi import HTTPException
from lazyops.libs import lazyload
from lazyops.libs.proxyobj import ProxyObject
from lazyops.utils.helpers import timed_cache

from ..types.errors import InvalidOperationException
from ..types.auth import AuthZeroTokenAuth
from ..types.clients import AuthZeroClientObject
from ..utils.lazy import get_az_settings, logger
from .tokens import ClientCredentialsFlow

from typing import Optional, List, Dict, Any, Union

if lazyload.TYPE_CHECKING:
    import niquests
    from niquests import Session, AsyncSession, Response, AsyncResponse
else:
    niquests = lazyload.LazyLoad("niquests")


class AZManagementAPI(ABC):
    
    """
    The Auth Zero Management API
    """

    if lazyload.TYPE_CHECKING:
        user_id: Optional[str] = None

    def __init__(
        self,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        oauth_url: Optional[str] = None,
        audience: Optional[str] = None,
        management_api_url: Optional[str] = None,
        use_http_session: Optional[bool] = False,
        **kwargs,
    ):
        """
        Initializes the Auth Zero Management API
        """
        self.settings = get_az_settings()
        self.management_api_url = management_api_url or self.settings.management_api_url
        self.token_flow = ClientCredentialsFlow(
            audience = audience or self.management_api_url,
            client_id = client_id,
            client_secret = client_secret,
            oauth_url = oauth_url,
            **kwargs,
        )

        self.auth = AuthZeroTokenAuth(token_flow=self.token_flow)
        self.use_http_session = use_http_session
        self._session: Optional['Session'] = None
        self._asession: Optional['AsyncSession'] = None

    def get_url(self, path: str) -> str:
        """
        Returns the URL
        """
        # if path.startswith('http'): return path
        return urljoin(self.management_api_url, path)
    
    def hget(self, path: str, **kwargs) -> 'Response':
        """
        Returns the Response
        """
        if not self.use_http_session:
            return niquests.get(self.get_url(path), auth = self.auth, **kwargs)
        return self.session.get(self.get_url(path), auth = self.auth, **kwargs)
    
    async def ahget(self, path: str, **kwargs) -> 'AsyncResponse':
        """
        Returns the Response
        """
        if self.use_http_session:
            return await self.asession.get(self.get_url(path), auth = self.auth, **kwargs)
        async with niquests.AsyncSession() as s:
            return await s.get(self.get_url(path), auth = self.auth, **kwargs)
    
    
    def hpost(self, path: str, **kwargs) -> 'Response':
        """
        Returns the Response
        """
        if not self.use_http_session:
            return niquests.post(self.get_url(path), auth = self.auth, **kwargs)
        return self.session.post(self.get_url(path), auth = self.auth, **kwargs)
    
    async def ahpost(self, path: str, **kwargs) -> 'AsyncResponse':
        """
        Returns the Response
        """
        if self.use_http_session:
            return await self.asession.post(self.get_url(path), auth = self.auth, **kwargs)
        async with niquests.AsyncSession() as s:
            return await s.post(self.get_url(path), auth = self.auth, **kwargs)

    def hput(self, path: str, **kwargs) -> 'Response':
        """
        Returns the Response
        """
        if not self.use_http_session:
            return niquests.put(self.get_url(path), auth = self.auth, **kwargs)
        return self.session.put(self.get_url(path), auth = self.auth, **kwargs)
    
    async def ahput(self, path: str, **kwargs) -> 'AsyncResponse':
        """
        Returns the Response
        """
        if self.use_http_session:
            return await self.asession.put(self.get_url(path), auth = self.auth, **kwargs)
        async with niquests.AsyncSession() as s:
            return await s.put(self.get_url(path), auth = self.auth, **kwargs)
        
    
    def hpatch(self, path: str, **kwargs) -> 'Response':
        """
        Returns the Response
        """
        if not self.use_http_session:
            return niquests.patch(self.get_url(path), auth = self.auth, **kwargs)
        return self.session.patch(self.get_url(path), auth = self.auth, **kwargs)
    
    async def ahpatch(self, path: str, **kwargs) -> 'AsyncResponse':
        """
        Returns the Response
        """
        if self.use_http_session:
            return await self.asession.patch(self.get_url(path), auth = self.auth, **kwargs)
        async with niquests.AsyncSession() as s:
            return await s.patch(self.get_url(path), auth = self.auth, **kwargs)

    def _validate_response(self, response: Union['Response', 'AsyncResponse']):
        """
        Validates the Response
        """
        try:
            response.raise_for_status()
        except niquests.HTTPError as e:
            if e.response.status_code in {400, 404}:
                raise HTTPException(status_code = e.response.status_code, detail = e.response.text) from e
            operation_name = f'{self.__class__.__name__}.{inspect.stack()[1].function}'
            msg = f'[{response.status_code}] Error doing `{operation_name}`'
            if hasattr(self, 'user_id'):
                msg += f' on user {self.user_id}'
            msg += f': {e.response.text}'
            logger.error(msg)
            raise InvalidOperationException(detail = msg) from e


    def _process_response(self, response: 'Response') -> Optional[Union[List, Dict]]:
        # sourcery skip: extract-method
        """
        Processes the Response
        """
        self._validate_response(response)
        if response.content: return response.json()
    
    def _404_to_empty_list(self, response: Union['Response', 'AsyncResponse']) -> Optional[List]:
        """
        Handles the Response
        """
        try:
            return self._process_response(response)
        except HTTPException as e:
            if e.status_code == 404: return []
            raise e


    @property
    def session(self) -> 'Session':
        """
        Returns the Session
        """
        if self._session is None:
            self._session = niquests.Session(
                pool_connections = 10,
                pool_maxsize = 10,
                retries = 5,
            )
        return self._session
    
    @property
    def asession(self) -> 'AsyncSession':
        """
        Returns the Async Session
        """
        if self._asession is None:
            self._asession = niquests.AsyncSession(
                pool_connections = 10,
                pool_maxsize = 10,
                retries = 5,
            )
        return self._asession
    

    """
    API Methods
    """

    def get_az_client(
        self,
        client_id: str,
        **kwargs
    ) -> Optional['AuthZeroClientObject']:
        """
        Returns the AuthZero Client
        """
        response = self.hget(f'clients/{client_id}')
        if response.status_code == 200:
            return AuthZeroClientObject.model_validate(response.json())
        logger.warning(f'[{response.status_code}] Error getting client: `{client_id}` {response.text}')
        return None
    

    async def aget_az_client(
        self,
        client_id: str,
        **kwargs
    ) -> Optional['AuthZeroClientObject']:
        """
        Returns the AuthZero Client
        """
        response = await self.ahget(f'clients/{client_id}')
        if response.status_code == 200:
            return AuthZeroClientObject.model_validate(response.json())
        logger.warning(f'[{response.status_code}] Error getting client: `{client_id}` {response.text}')
        return None
    
    @timed_cache(600)
    def get_service_client_name(
        self,
        client_id: str,
    ) -> Optional[str]:
        """
        Returns the Service Client Name
        """
        client = self.get_az_client(client_id = client_id)
        return None if client is None else client.name

    @timed_cache(600)
    async def aget_service_client_name(
        self,
        client_id: str,
    ) -> Optional[str]:
        """
        Returns the Service Client Name
        """
        client = await self.aget_az_client(client_id = client_id)
        return None if client is None else client.name



AZManagementClient: AZManagementAPI = ProxyObject(
    AZManagementAPI,
    obj_kwargs = {'use_http_session': False},
)


from __future__ import annotations

"""
The PostHog Client

- Uses niquest 
"""

# WARNING: This is a WIP and not yet fully tested.


import abc
import time
import asyncio
import functools
from lazyops.imports._niquests import resolve_niquests
resolve_niquests(True)

import niquests
from lazyops.libs.pooler import ThreadPooler
from lazyops.utils.logs import logger, null_logger, Logger
from lazyops.utils.times import Timer
from typing import Optional, Dict, Any, List, Union, Type, Set, Tuple, Callable, TypeVar, TYPE_CHECKING

from .config import PostHogSettings
from .utils import get_posthog_settings, register_posthog_client, get_posthog_client, has_existing_posthog_client
from .types import PostHogAuth, PostHogEndpoint, EventQueue, EventT

if TYPE_CHECKING:
    from .config import PostHogSettings



RT = TypeVar('RT')



class PostHogClient(abc.ABC):
    """
    An Asyncronous PostHog Client that utilizes Intelligent Batching and Asynchronous Task Queues
    to send events to PostHog in a more efficient and scalable manner.

    This client is designed with the following features in mind:
    - Intelligent Batching: The client automatically batches events based on the configured batch size and sends them in batches once the batch size is reached or the maximum batch interval has been reached.
    - Asynchronous Task Queues: The client uses asynchronous task queues to handle the sending of events. This allows the client to send events in parallel and improves the overall performance of the client.
    - Context Manager: The client provides a context manager that can be used to force sending of events immediately after exiting the context rather than waiting for the batching logic to complete.
    - Non-Blocking: By utilizing asynchronous task queues and non-blocking requests, the client can handle a large number of events without blocking the main thread.
    - Lazily Initializes: The client lazily initializes the PostHog client and only initializes it when the first event is added.
    """

    settings: Optional['PostHogSettings'] = None

    def __init__(
        self,
        endpoint: Optional[str] = None,
        enabled: Optional[bool] = None,
        api_key: Optional[str] = None,
        project_id: Optional[str] = None,
        client_timeout: Optional[float] = None,
        batch_size: Optional[int] = None, # If the total events is greater than this, it will dispatch the batch
        batch_interval: Optional[float] = None, # if the wait duration is greater than this, it will dispatch the batch
        debug_enabled: Optional[bool] = None,
        default_retries: Optional[int] = None,
        batched: Optional[bool] = None,
        num_workers: Optional[int] = None,
        upkeep_interval: Optional[float] = 0.1,
        dryrun: Optional[bool] = None,
        **kwargs
    ):
        """
        Initializes the PostHog Client

        - `endpoint`: The endpoint to use for the PostHog API
        - `enabled`: Whether the client is enabled or not
        - `api_key`: The API key to use for authentication
        - `project_id`: The project ID to use for the PostHog API
        - `client_timeout`: The timeout to use for the client
        - `batch_size`: The maximum number of events to batch before sending
        - `batch_interval`: The maximum time to wait before sending the batch
        - `debug_enabled`: Whether to enable debug logging
        - `default_retries`: The number of retries to attempt before giving up
        - `batched`: Whether to batch events before sending
        - `num_workers`: The number of workers to use for the event queue
        - `upkeep_interval`: The interval to check for events and send them
        - `dryrun`: Whether to run in dryrun mode. This will not actually send any events to PostHog

        These parameters will override the values in the PosthogSettings
        """
        # Reuse the existing client if it exists
        # which will prevent multiple clients from being created
        self.settings = get_posthog_settings()
        # if has_existing_posthog_client():
        #     self = get_posthog_client(
        #         endpoint = endpoint,
        #         enabled = enabled,
        #         api_key = api_key,
        #         project_id = project_id,
        #         client_timeout = client_timeout,
        #         batch_size = batch_size,
        #         batch_interval = batch_interval,
        #         debug_enabled = debug_enabled,
        #         default_retries = default_retries,
        #         batched = batched,
        #         num_workers = num_workers,
        #         **kwargs
        #     )
        #     return
        self.upkeep_interval = upkeep_interval or 0.1
        self.dryrun = dryrun
        self._kwargs = kwargs
        self.configure(
            endpoint = endpoint,
            enabled = enabled,
            api_key = api_key,
            project_id = project_id,
            client_timeout = client_timeout,
            batch_size = batch_size,
            batch_interval = batch_interval,
            debug_enabled = debug_enabled,
            default_retries = default_retries,
            batched = batched,
            num_workers = num_workers,
        )

        self.endpoint = PostHogEndpoint(endpoint = self.settings.endpoint)
        self.auth = PostHogAuth(api_key = self.settings.api_key)

        self.logger = logger if (self.settings.debug_enabled is None or self.settings.debug_enabled) else null_logger
        self.pooler = ThreadPooler

        self._loop: Optional[asyncio.AbstractEventLoop] = None
        self._session: Optional['niquests.Session'] = None
        self._asession: Optional['niquests.AsyncSession'] = None

        self.main_task: Optional[asyncio.Task] = None
        self.event: Optional[asyncio.Event] = None
        self.started_time: Optional[Timer] = None
        self.event_queue: Optional[EventQueue] = None
        self.events_sent: Optional[int] = 0
        self.send_duration: Optional[float] = 0.0
        self.should_send_ctx: Optional[bool] = None

        self.successful_events: Optional[int] = 0
        self.failed_events: Optional[int] = 0

        self.tasks: Optional[Set[asyncio.Task]] = set()
        self._task_queue: Optional[asyncio.Queue] = None

        self._lock: Optional[asyncio.Lock] = None
        self._kwargs = kwargs
        register_posthog_client(self, **kwargs)


    def configure(
        self,
        endpoint: Optional[str] = None,
        enabled: Optional[bool] = None,
        api_key: Optional[str] = None,
        project_id: Optional[str] = None,
        client_timeout: Optional[float] = None,
        batch_size: Optional[int] = None, # If the total events is greater than this, it will dispatch the batch
        batch_interval: Optional[float] = None, # if the wait duration is greater than this, it will dispatch the batch
        debug_enabled: Optional[bool] = None,
        default_retries: Optional[int] = None,
        batched: Optional[bool] = None,
        num_workers: Optional[int] = None,
        upkeep_interval: Optional[float] = 0.1,
        dryrun: Optional[bool] = None,
        **kwargs
    ):
        """
        Configures the PostHog Client
        """
        if endpoint is not None: self.settings.endpoint = endpoint
        if enabled is not None: self.settings.enabled = enabled
        if api_key is not None: self.settings.api_key = api_key
        if project_id is not None: self.settings.project_id = project_id
        if client_timeout is not None: self.settings.client_timeout = client_timeout
        if batch_size is not None: self.settings.batch_size = batch_size
        if batch_interval is not None: self.settings.batch_interval = batch_interval
        if debug_enabled is not None: self.settings.debug_enabled = debug_enabled
        if batched is not None: self.settings.batched = batched
        if num_workers is not None: self.settings.num_workers = num_workers
        if default_retries is not None: self.settings.default_retries = default_retries
        if upkeep_interval is not None: self.upkeep_interval = upkeep_interval
        if dryrun is not None: self.dryrun = dryrun
        if kwargs: self._kwargs.update(kwargs)
        self.settings.update_enabled()
        return self

    @property
    def autologger(self) -> 'Logger':
        """
        Returns the autologger
        """
        return logger if self.settings.debug_enabled else null_logger

    @property
    def started(self) -> bool:
        """
        Returns True if the client is started
        """
        return self.main_task is not None
    
    @property
    def enabled(self) -> bool:
        """
        Returns True if the client is enabled
        """
        return self.settings._enabled
    
    def get_session_kwargs(self, **kwargs) -> Dict[str, Any]:
        """
        Returns the session kwargs
        """
        return {
            'pool_connections': self._kwargs.get('pool_connections', 10),
            'pool_maxsize': self._kwargs.get('pool_maxsize', 10),
            'retries': self._kwargs.get('retries', self.settings.default_retries),
            'resolver': self._kwargs.get('resolver', 'doh+cloudflare://'),
            'multiplexed': True,
            **kwargs,
        }


    @property
    def session(self) -> 'niquests.Session':
        """
        Returns the Session
        """
        if self._session is None: self._session = niquests.Session(**self.get_session_kwargs())
        return self._session
    
    @property
    def asession(self) -> 'niquests.AsyncSession':
        """
        Returns the Async Session
        """
        if self._asession is None: self._asession = niquests.AsyncSession(**self.get_session_kwargs())
        return self._asession
    
    @property
    def task_queue(self) -> asyncio.Queue:
        """
        Returns the Task Queue
        """
        if self._task_queue is None: self._task_queue = asyncio.Queue()
        return self._task_queue
    
    @property
    def lock(self) -> asyncio.Lock:
        """
        Returns the Lock
        """
        if self._lock is None: self._lock = asyncio.Lock()
        return self._lock

    def reset_session(self):
        """
        Resets the session
        """
        if self._session is not None: self._session.close()
        self._session = None
    
    async def areset_session(self):
        """
        Resets the async session
        """
        if self._asession is not None: await self._asession.close()
        self._asession = None
        if self._session is not None:  self._session.close()
        self._session = None

    def add_event(self, event: Union[EventT, str], **kwargs):
        """
        Adds the event to the queue

        - `event`: The event to add to the queue
        - `kwargs`: The kwargs to add to the event

        The event can be a string or a PostHog event object.
        """
        if not self.enabled: return
        if not self.started: self.start_task_queue()
        
        self.task_queue.put_nowait((event, kwargs))

    async def send_events(
        self, 
        force: Optional[bool] = None,
        batched: Optional[bool] = None,
    ) -> Tuple[List[niquests.Response], int]:  # sourcery skip: move-assign
        """
        Sends the events
        """
        if not self.event_queue: 
            self.autologger.info('Empty Queue. Skipping Send', colored = True)
            return None, None
        if self.lock.locked() and not force: 
            self.autologger.info('Locked. Skipping Send', colored = True)
            return None, None
        responses = []
        sent = 0 
        batched = batched if batched is not None else self.settings.batched
        async with self.lock:
            async with self.asession as s:
                req_func = functools.partial(s.post, timeout = self.settings.client_timeout, auth = self.auth, stream = True)
                base_json = {'api_key': self.settings.api_key}
                if self.event_queue.capture_events:
                    events = self.event_queue.prepare_events(kind = 'capture', batched = batched, exclude_none = True, clear_after = True)
                    sent += len(events)
                    if batched: 
                        if self.dryrun:
                            self.autologger.info(f'[DRYRUN] (Force: {force}, In Context: {self.should_send_ctx}) Capture Batched Event: {events} ', prefix = f'{self.endpoint.batch}', colored = True)
                        else: responses.append(await req_func(self.endpoint.batch, json = {'batch': events, **base_json}))
                    else:
                        for event in events:
                            if self.dryrun:
                                self.autologger.info(f'[DRYRUN] (Force: {force}, In Context: {self.should_send_ctx}) Capture Event: {event}', prefix = f'{self.endpoint.capture}', colored = True)
                                continue
                            responses.append(await req_func(self.endpoint.capture, json = {**event, **base_json}))
                    
                
                # TODO: add identify ?
                await s.gather(*responses)
            self.event_queue.clear()
        self.events_sent += sent
        self.should_send_ctx = False
        return responses, sent

    async def handle_event(self, event, **kwargs):
        """
        Handles the event
        """
        self.event_queue.add_event(event, **kwargs)
        self.autologger.info(f'Added Event: {len(self.event_queue)}/{self.settings.batch_size}', prefix = 'PostHog', colored = True)


    async def should_send_events(self, ts: Optional[float] = None, force: Optional[bool] = None, **kwargs) -> bool:
        """
        Checks if the events should be sent
        """
        if force: return True
        if self.should_send_ctx: return True
        if ts and (elapsed_s := (time.monotonic() - ts)) > self.settings.batch_interval: 
            self.autologger.info(f'Sending Batch: {len(self.event_queue)} @ {elapsed_s:.2f}s', prefix = 'Max Interval')
            return True

        if self.settings.batch_size is None: return False
        if len(self.event_queue) >= self.settings.batch_size:
            self.autologger.info(f'Sending Batch: {len(self.event_queue)}/{self.settings.batch_size} ', prefix = 'Max Size')
            return True
        return False

    async def handle_responses(self, batch: Tuple[List[niquests.Response], int], t: Timer, worker_n: int):
        """
        Handles the responses
        """
        responses, sent_events = batch
        if sent_events: self.send_duration += t.duration
        if not responses: return
        for response in responses:
            if response.status_code != 200:
                self.failed_events += 1
                self.autologger.info(f'[{response.status_code}] Error Sending Event: |y|{await response.text}|e|', prefix = f'PostHog Worker: {worker_n}', colored = True) 
            else: self.successful_events += 1

    async def upkeep(self, worker_n: Optional[int] = None):
        """
        This is a background task to handle upkeep
        """
        
        self.autologger.info('Upkeep Started', prefix = f'PostHog Worker: {worker_n}', colored = True)
        wt = Timer()
        ts = time.monotonic()
        while not self.event.is_set():
            try:
                item = self.task_queue.get_nowait()
                event, kwargs = item
                await self.handle_event(event, **kwargs)
                if await self.should_send_events(ts):
                    tt = Timer(format_ms = True, format_short = 2)
                    response_batch = await self.send_events()
                    await self.handle_responses(response_batch, tt, worker_n)
                    ts = time.monotonic()
                self.task_queue.task_done()

            except asyncio.QueueEmpty: await asyncio.sleep(self.upkeep_interval)
            except (Exception, asyncio.CancelledError) as e:
                if self.event.is_set(): 
                    self.autologger.info(f'Event Set. Exiting after {wt.total_s} Uptime', prefix = f'PostHog Worker: {worker_n}', colored = True)
                    return
                self.autologger.info(f'Error in Upkeep: |r|{e}|e|. Exiting after {wt.total_s} Uptime', prefix = f'PostHog Worker: {worker_n}', colored = True)
                return
            
            

    async def run_task_queue(self, num_workers: Optional[int] = None):
        """
        Runs the Task Queue
        """
        self.event = asyncio.Event()
        num_workers = num_workers or self.settings.num_workers
        try:
            for n in range(num_workers):
                self.tasks.add(asyncio.create_task(self.upkeep(n)))
            await self.event.wait()
        except Exception as e:
            logger.error(f'Error in PostHog Client: {e}')
        finally:
            await self.stop_task_queue()
            self.tasks.remove(self.main_task)
            self.main_task = None

    def get_stat_message(self) -> str:
        """
        Returns the stat message
        """
        m = f'Total Uptime: |g|{self.started_time.total_s}|e|. Events Sent: |g|{self.events_sent}|e|. Total Request Duration: |g|{self.started_time.pformat_duration(self.send_duration, short = 2, include_ms = True)}|e|.'
        if self.send_duration: m += f' Avg |g|{self.events_sent / self.send_duration:.2f} events/s|e|'
        return m


    async def stop_task_queue(self):
        """
        Stops the Task Queue
        """
        if self.main_task is None: return
        self.autologger.info('Stopping PostHog Event Queue')
        self.event.set()
        await self.send_events(force = True)
        all_tasks = list(self.tasks)
        self.tasks.clear()
        for task in all_tasks:
            if not task.done(): task.cancel()
        await self.areset_session()
        self.autologger.info(f'Event Queue Stopped. {self.get_stat_message()}', prefix = 'PostHog', colored = True)
        if self.main_task: self.main_task.cancel()
        await asyncio.gather(*all_tasks, return_exceptions=True)

    def start_task_queue(self, num_workers: Optional[int] = None):
        """
        Starts the Task Queue
        """
        if self.started: return
        if not self.enabled: 
            self.logger.warning('PostHog is not enabled. Please set `POSTHOG_API_KEY` to enable PostHog')
            return
        num_workers = num_workers or self.settings.num_workers
        self.autologger.info(f'Starting PostHog Event Queue with |g|{num_workers}|e| Workers', colored = True)
        self.started_time = Timer()
        self.event_queue = EventQueue()
        self.main_task = asyncio.create_task(self.run_task_queue(num_workers = num_workers))
        
    def __enter__(self):
        """
        Context Manager for the PostHog Client

        This is used to ensure that any events that are added during the context manager are sent immediately
        """
        self.start_task_queue()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Context Manager for the PostHog Client
        """
        self.should_send_ctx = True
    
    async def __aenter__(self):
        """
        Async Context Manager for the PostHog Client

        This is used to ensure that any events that are added during the context manager are sent immediately
        """
        self.start_task_queue()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
        Async Context Manager for the PostHog Client
        """
        self.should_send_ctx = True
    

    def capture(
        self,
        event: Optional[str] = None,
        distinct_id: Optional[str] = None,
        properties: Optional[Dict[str, Any]] = None,
        **_kwargs: Any, 
    ) -> Callable[..., RT]:
        """
        Creates a decorator that can be used to capture events
        """
        def decorator(func: Callable[..., RT]) -> Callable[..., RT]:
            nonlocal event, distinct_id, properties
            if event is None: event = func.__name__
            properties = properties or {}
            ph_ctx = {'event': event, 'distinct_id': distinct_id, 'properties': properties, **_kwargs}

            if ThreadPooler.is_coro(func):
                @functools.wraps(func)
                async def capture_decorator(*args, **kwargs):
                    result = await func(*args, ph_ctx = ph_ctx, **kwargs)
                    self.add_event(**ph_ctx)
                    return result
            else:
                @functools.wraps(func)
                def capture_decorator(*args, **kwargs):
                    result = func(*args, ph_ctx = ph_ctx, **kwargs)
                    self.add_event(**ph_ctx)
                    return result
            return capture_decorator
        
        return decorator
            

    def flush(self, force: Optional[bool] = True) -> None:
        """
        Flushes the events
        """
        self.tasks.add(
            asyncio.create_task(self.send_events(force = force))
        )

    async def aflush(self, force: Optional[bool] = True) -> None:
        """
        Flushes the events
        """
        await self.send_events(force = force)


from __future__ import annotations

"""
Credits:

https://gist.github.com/agronholm/458637aa569720fb1305cc74347e3e1d
"""

__all__ = ["connect", "Connection", "Cursor"]

import sqlite3
import contextlib
from collections.abc import Callable, Sequence
from functools import partial, wraps
from typing import Any, Literal, Optional, overload

from anyio import CapacityLimiter, to_thread


class Connection:
    def __init__(self, _real_connection: sqlite3.Connection, limit: Optional[int] = None) -> None:
        self._real_connection = _real_connection
        limit = limit or 1
        self._limiter = CapacityLimiter(limit)

    @wraps(sqlite3.Connection.close)
    async def close(self):
        return await to_thread.run_sync(self._real_connection.close, limiter=self._limiter)

    @wraps(sqlite3.Connection.commit)
    async def commit(self):
        return await to_thread.run_sync(self._real_connection.commit, limiter=self._limiter)

    @wraps(sqlite3.Connection.rollback)
    async def rollback(self):
        return await to_thread.run_sync(self._real_connection.rollback, limiter=self._limiter)

    async def cursor(self, factory: Callable[[sqlite3.Connection], sqlite3.Cursor] = sqlite3.Cursor) -> Cursor:
        real_cursor = await to_thread.run_sync(self._real_connection.cursor, factory, limiter=self._limiter)
        return Cursor(real_cursor, self._limiter)

    async def __aenter__(self) -> "Connection":
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.close()


class Cursor:
    def __init__(self, real_cursor: sqlite3.Cursor, limiter: CapacityLimiter) -> None:
        self._real_cursor = real_cursor
        self._limiter = limiter

    @property
    def description(self) -> Any:
        return self._real_cursor.description

    @property
    def rowcount(self) -> int:
        return self._real_cursor.rowcount

    @property
    def arraysize(self) -> int:
        return self._real_cursor.arraysize
    
    @property
    def lastrowid(self) -> int:
        return self._real_cursor.lastrowid

    @wraps(sqlite3.Cursor.close)
    async def close(self) -> None:
        await to_thread.run_sync(self._real_cursor.close, limiter=self._limiter)

    @wraps(sqlite3.Cursor.execute)
    async def execute(self, sql: str, parameters: Sequence[Any] = (), /) -> Cursor:
        real_cursor = await to_thread.run_sync(self._real_cursor.execute, sql, parameters, limiter=self._limiter)
        return Cursor(real_cursor, self._limiter)

    @wraps(sqlite3.Cursor.executemany)
    async def executemany(self, sql: str, parameters: Sequence[Any], /) -> Cursor:
        real_cursor = await to_thread.run_sync(self._real_cursor.executemany, sql, parameters, limiter=self._limiter)
        return Cursor(real_cursor, self._limiter)

    @wraps(sqlite3.Cursor.executescript)
    async def executescript(self, sql_script: str, /) -> Cursor:
        real_cursor = await to_thread.run_sync(self._real_cursor.executescript, sql_script, limiter=self._limiter)
        return Cursor(real_cursor, self._limiter)

    @wraps(sqlite3.Cursor.fetchone)
    async def fetchone(self) -> tuple[Any, ...] | None:
        return await to_thread.run_sync(self._real_cursor.fetchone, limiter=self._limiter)

    @wraps(sqlite3.Cursor.fetchmany)
    async def fetchmany(self, size: int) -> list[tuple[Any, ...]]:
        return await to_thread.run_sync(self._real_cursor.fetchmany, size, limiter=self._limiter)

    @wraps(sqlite3.Cursor.fetchall)
    async def fetchall(self) -> list[tuple[Any, ...]]:
        return await to_thread.run_sync(self._real_cursor.fetchall, limiter=self._limiter)


    async def __aenter__(self) -> "Cursor":
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.close()

@overload
async def connect(
    database: str,
    timeout: Optional[float] = None,
    detect_types: Optional[int] = None,
    isolation_level: Literal['DEFERRED', 'IMMEDIATE', 'EXCLUSIVE'] = None,
    cached_statements: Optional[int] = 128,
    uri: Optional[bool] = False,
    limit: Optional[int] = None,
) -> Connection:
    """
    Connects to the database
    """
    ...


async def connect(
    database: str,
    limit: Optional[int] = None,
    **kwargs,
) -> Connection:
    """
    Connects to the database
    """
    _ = kwargs.pop('check_same_thread', None)
    real_connection = await to_thread.run_sync(
        partial(sqlite3.connect, database, check_same_thread=False, **kwargs)
    )
    return Connection(real_connection, limit = limit)


from __future__ import annotations

"""
Base API Client and HTTP Pool Client
"""

import asyncio
import random
import inspect
import aiohttpx
import functools
import subprocess
from pydantic import BaseModel
from urllib.parse import urlparse
from lazyops.libs.proxyobj import ProxyObject, proxied
from .base import BaseGlobalClient, cachify
from .utils import aget_root_domain, get_user_agent, http_retry_wrapper

from typing import Optional, Type, TypeVar, Literal, Union, Set, Awaitable, Any, Dict, List, Callable, overload, TYPE_CHECKING

if TYPE_CHECKING:
    from kvdb import KVDBSession, SerializerT
    from docx.document import Document


ReturnType = Literal['json', 'text', 'bytes']
ReturnModelT = TypeVar('ReturnModelT', bound = BaseModel)
ReturnTypeT = TypeVar('ReturnTypeT', bound = Union[str, bytes, Dict[str, Any], List[Any], BaseModel, List[BaseModel], Dict[str, BaseModel], List[Dict[str, BaseModel]]])


@cachify.register_object()
class BaseAPIClient(BaseGlobalClient):

    name: Optional[str] = 'api'
    verbose_errors: Optional[bool] = True
    http_timeout: Optional[float] = None

    _endpoint: Optional[str] = None
    _headers: Optional[Dict[str, Any]] = None
    _api: Optional[aiohttpx.Client] = None


    @property
    def endpoint(self) -> str:
        """
        Returns the endpoint

        - Override this to add custom endpoints
        """
        return self._endpoint
    
    def get_user_agent(self, *args, **kwargs) -> str:
        """
        Returns the user agent
        """
        return get_user_agent(*args, **kwargs)
    
    @property
    def headers(self) -> Dict[str, Any]:
        """
        Returns the headers

        - Override this to add custom headers
        """
        return self._headers or {}
    
    @property
    def api_client_kwargs(self) -> Dict[str, Any]:
        """
        Returns the api client kwargs

        - Override this to add custom kwargs
        """
        return {}
    

    def cachify_get_exclude_keys(self, func: str, **kwargs) -> List[str]:
        """
        Gets the exclude keys
        """
        return [
            'background',
            'callback',
            'retryable',
            'retry_limit',
            'validate_url',
            'cachable',
            'disable_cache',
            'overwrite_cache',
        ]
    
    def configure_api_client(self, *args, **kwargs) -> aiohttpx.Client:
        """
        Configures the API Client
        """
        if hasattr(self.settings, 'clients') and hasattr(self.settings.clients, 'http_pool'):
            limits = aiohttpx.Limits(
                max_connections = self.settings.clients.http_pool.max_connections,
                max_keepalive_connections = self.settings.clients.http_pool.max_keepalive_connections,
                keepalive_expiry = self.settings.clients.http_pool.keepalive_expiry,
            )
            timeout = self.http_timeout or self.settings.clients.http_pool.default_timeout
        else:
            limits = aiohttpx.Limits(
                max_connections = 100,
                max_keepalive_connections = 20,
                keepalive_expiry = 60,
            )
            timeout = self.http_timeout or 60
        return aiohttpx.Client(
            base_url = self.endpoint,
            limits = limits,
            timeout = timeout,
            headers = self.headers,
            verify = False,
            **self.api_client_kwargs,
        )
    

    @property
    def api(self) -> aiohttpx.Client:
        """
        Returns the API Client
        """
        if self._api is None:
            self._api = self.configure_api_client()
        return self._api
    

    async def areset_api(self):
        """
        Resets the api client
        """
        if self._api:
            await self._api.aclose()
            self._api = None

    def wrap_retryable_method(
        self,
        func: Callable[..., Any],
        retry_limit: Optional[int] = 3,
    ) -> Union[Callable[..., aiohttpx.Response], Callable[..., Awaitable[aiohttpx.Response]]]:
        """
        Wraps a retryable method
        """
        if isinstance(func, str): func = getattr(self.api, func)
        return http_retry_wrapper(max_tries = retry_limit + 1)(func)
    
    def cachify_validator_is_disabled(self, *args, disable_cache: Optional[bool] = None, **kwargs) -> bool:
        """
        Checks if the function is disabled
        """
        return disable_cache
    
    def cachify_validator_is_overwrite(self, *args, overwrite_cache: Optional[bool] = None, **kwargs) -> bool:
        """
        Checks if the function is overwrite
        """
        return overwrite_cache
    

    def cachify_get_name_builder_kwargs(self, func: str, **kwargs) -> Dict[str, Any]:
        """
        Gets the name builder kwargs
        """
        return {
            'include_http_methods': True,
        }

    """
    Modified HTTP Methods
    """

    @overload
    def get(
        self, 
        url: str, 
        *args,
        return_type: Optional[Union[ReturnType, ReturnModelT]] = 'json', 
        cachable: Optional[bool] = True,
        overwrite_cache: Optional[bool] = None,
        disable_cache: Optional[bool] = None,
        background: Optional[bool] = False,
        callback: Optional[Callable[..., Any]] = None,
        retryable: Optional[bool] = False,
        retry_limit: Optional[int] = 3,
        **kwargs
    ) -> ReturnTypeT:
        ...

    @cachify.register()
    def get(
        self, 
        url: str, 
        *args,
        return_type: Optional[Union[ReturnType, ReturnModelT]] = 'json', 
        cachable: Optional[bool] = True,
        overwrite_cache: Optional[bool] = None,
        disable_cache: Optional[bool] = None,
        background: Optional[bool] = False,
        callback: Optional[Callable[..., Any]] = None,
        retryable: Optional[bool] = False,
        retry_limit: Optional[int] = 3,
        **kwargs
    ) -> ReturnTypeT:
        """
        Creates a GET Request and returns a `ReturnTypeT`
        """
        if background:
            return self.pooler.create_background(
                self._get,
                url = url,
                *args,
                return_type = return_type,
                task_callback = callback,
                retryable = retryable,
                retry_limit = retry_limit,
                **kwargs
            )
        return self._get(url = url, *args, return_type = return_type, retryable = retryable, retry_limit = retry_limit, **kwargs)
    
    @cachify.register()
    def post(
        self, 
        url: str, 
        *args,
        return_type: Optional[Union[ReturnType, ReturnModelT]] = 'json', 
        cachable: Optional[bool] = True,
        overwrite_cache: Optional[bool] = None,
        disable_cache: Optional[bool] = None,
        background: Optional[bool] = False,
        callback: Optional[Callable[..., Any]] = None,
        retryable: Optional[bool] = False,
        retry_limit: Optional[int] = 3,
        **kwargs
    ) -> ReturnTypeT:
        """
        Creates a POST Request and returns a `ReturnTypeT`
        """
        if background:
            return self.pooler.create_background(
                self._post,
                url = url,
                *args,
                return_type = return_type,
                task_callback = callback,
                retryable = retryable,
                retry_limit = retry_limit,
                **kwargs
            )
        return self._post(url = url, *args, return_type = return_type, retryable = retryable, retry_limit = retry_limit, **kwargs)
    
    @overload
    async def aget(
        self,
        url: str, 
        *args,
        return_type: Optional[Union[ReturnType, ReturnModelT]] = 'json', 
        cachable: Optional[bool] = True,
        overwrite_cache: Optional[bool] = None,
        disable_cache: Optional[bool] = None,
        background: Optional[bool] = False,
        callback: Optional[Callable[..., Any]] = None,
        retryable: Optional[bool] = False,
        retry_limit: Optional[int] = 3,
        **kwargs
    ) -> ReturnTypeT:
        """
        Creates a GET Request and returns a `ReturnTypeT`
        """

    @cachify.register()
    async def aget(
        self, 
        url: str, 
        *args,
        return_type: Optional[Union[ReturnType, ReturnModelT]] = 'json', 
        cachable: Optional[bool] = True,
        overwrite_cache: Optional[bool] = None,
        disable_cache: Optional[bool] = None,
        background: Optional[bool] = False,
        callback: Optional[Callable[..., Any]] = None,
        retryable: Optional[bool] = False,
        retry_limit: Optional[int] = 3,
        **kwargs
    ) -> ReturnTypeT:
        """
        Creates a GET Request and returns a `ReturnTypeT`
        """
        if background:
            return self.pooler.create_background(
                self._aget,
                url = url,
                *args,
                return_type = return_type,
                task_callback = callback,
                retryable = retryable,
                retry_limit = retry_limit,
                **kwargs
            )
        return await self._aget(url = url, *args, return_type = return_type, retryable = retryable, retry_limit = retry_limit,  **kwargs)
    
    @cachify.register()
    async def apost(
        self, 
        url: str, 
        *args,
        return_type: Optional[Union[ReturnType, ReturnModelT]] = 'json', 
        cachable: Optional[bool] = True,
        overwrite_cache: Optional[bool] = None,
        disable_cache: Optional[bool] = None,
        background: Optional[bool] = False,
        callback: Optional[Callable[..., Any]] = None,
        retryable: Optional[bool] = False,
        retry_limit: Optional[int] = 3,
        **kwargs
    ) -> ReturnTypeT:
        """
        Creates a POST Request and returns a `ReturnTypeT`
        """
        # base_post = self._apost_cachable if cachable else self._apost
        if background:
            return self.pooler.create_background(
                self._apost,
                url = url,
                *args,
                return_type = return_type,
                task_callback = callback,
                retryable = retryable,
                retry_limit = retry_limit,
                **kwargs
            )
        return await self._apost(url = url, *args, return_type = return_type, retryable = retryable, retry_limit = retry_limit, **kwargs)
    


    """
    Basic HTTP Methods
    """

    def _get_(self, url: str, **kwargs) -> aiohttpx.Response:
        """
        Makes a GET request
        """
        return self.api.get(url, **kwargs)
    
    def _post_(self, url: str, **kwargs) -> aiohttpx.Response:
        """
        Makes a POST request
        """
        return self.api.post(url, **kwargs)
    
    def _validate_url(self, url: str) -> Union[bool, str]:
        """
        Quickly validates a URL
        """
        try:
            response = self.api.head(url, follow_redirects = True)
            try:
                response.raise_for_status()
                return True
            except Exception as e:
                return f'[{response.status_code}]: {response.text}'
        except Exception as e:
            return f'[{type(e)}]: {str(e)}'
        
    def _fetch_content_type(self, url: str) -> Optional[str]:
        """
        Fetches the content type
        """
        try:
            response = self.api.head(url, follow_redirects = True)
            return response.headers.get('content-type')
        except Exception as e:
            return None
        
    async def _aget_(self, url: str, **kwargs) -> aiohttpx.Response:
        """
        Makes a GET request
        """
        return await self.api.async_get(url, **kwargs)
    
    async def _apost_(self, url: str, **kwargs) -> aiohttpx.Response:
        """
        Makes a POST request
        """
        return await self.api.async_post(url, **kwargs)
    
    async def _avalidate_url(self, url: str) -> Union[bool, str]:
        """
        Quickly validates a URL
        """
        try:
            response = await self.api.async_head(url, follow_redirects = True)
            try:
                response.raise_for_status()
                return True
            except Exception as e:
                return f'[{response.status_code}] {response.text}'
        except Exception as e:
            return f'[{type(e)}]: {str(e)}'
        
    async def _afetch_content_type(self, url: str) -> Optional[str]:
        """
        Fetches the content type
        """
        try:
            response = await self.api.async_head(url, follow_redirects = True)
            return response.headers.get('content-type')
        except Exception as e:
            return None

    """
    Enhanced HTTP Methods
    """

    def _get(
        self,
        url: str,
        *args,
        return_type: Optional[Union[ReturnType, ReturnModelT]] = 'json',
        retryable: Optional[bool] = False,
        retry_limit: Optional[int] = 3,
        **kwargs
    ) -> ReturnTypeT:
        """
        Creates a GET Request and returns a `ReturnTypeT`
        """
        get_func = self._get_
        if retryable: get_func = http_retry_wrapper(max_tries = retry_limit + 1)(get_func)
        response = get_func(url, *args, **kwargs)
        return self.handle_response(response, return_type = return_type)
    
    def _post(
        self,
        url: str,
        *args,
        return_type: Optional[Union[ReturnType, ReturnModelT]] = 'json',
        retryable: Optional[bool] = False,
        retry_limit: Optional[int] = 3,
        **kwargs
    ) -> ReturnTypeT:
        """
        Creates a POST Request and returns a `ReturnTypeT`
        """
        post_func = self._post_
        if retryable: post_func = http_retry_wrapper(max_tries = retry_limit + 1)(post_func)
        response = post_func(url, *args, **kwargs)
        return self.handle_response(response, return_type = return_type)


    async def _aget(
        self,
        url: str,
        *args,
        return_type: Optional[Union[ReturnType, ReturnModelT]] = 'json',
        retryable: Optional[bool] = False,
        retry_limit: Optional[int] = 3,
        **kwargs
    ) -> ReturnTypeT:
        """
        Creates a GET Request and returns a `ReturnTypeT`
        """
        get_func = self._aget_
        if retryable: get_func = http_retry_wrapper(max_tries = retry_limit + 1)(get_func)
        response = await get_func(url, *args, **kwargs)
        return self.handle_response(response, return_type = return_type)

    
    async def _apost(
        self,
        url: str,
        *args,
        return_type: Optional[Union[ReturnType, ReturnModelT]] = 'json',
        retryable: Optional[bool] = False,
        retry_limit: Optional[int] = 3,
        **kwargs
    ) -> ReturnTypeT:
        """
        Creates a POST Request and returns a `ReturnTypeT`
        """
        post_func = self._apost_
        if retryable: post_func = http_retry_wrapper(max_tries = retry_limit + 1)(post_func)
        response = await post_func(url, *args, **kwargs)
        return self.handle_response(response, return_type = return_type)
    

    def handle_response(
        self, 
        response: aiohttpx.Response, 
        return_type: Optional[Union[ReturnType, ReturnModelT]] = 'json',
    ) -> Union[ReturnModelT, Dict[str, Any]]:
        """
        Handle the response
        """
        try:
            response.raise_for_status()
            if not return_type: return response
            if not isinstance(return_type, str):
                # Must be a model
                return return_type.model_validate(response.json())
            return_type = str(return_type).lower()
            if return_type == 'json':
                return response.json()
            if return_type == 'text':
                return response.text
            if return_type == 'bytes':
                return response.content
            else:
                raise ValueError(f'Invalid return type: {return_type}')
        except aiohttpx.HTTPStatusError as e:
            if self.settings.is_production_env:
                self.logger.warning(f'[{response.status_code} - {e.request.url}]')
            else:
                self.logger.error(f'[{response.status_code} - {e.request.url}] {response.text}')
        except Exception as e:
            self.logger.trace(f'Error in response: {response.text}', e)
            raise e
    

class HTTPPoolClient(BaseAPIClient):
    """
    This is a Global HTTP API Client that can be used by any client
    """
    name: Optional[str] = 'http'
    cachify_ttl: Optional[int] = 60 * 60 * 24 * 7 # 1 Week

    google_csx_base_url: Optional[str] = 'https://www.googleapis.com/customsearch/v1'

    # Extended Methods
    _pdftotext_enabled: Optional[bool] = None
    _doctotext_enabled: Optional[bool] = None

    _google_csx_api_key: Optional[str] = None
    _google_csx_id: Optional[str] = None


    @property
    def google_csx_api_key(self) -> str:
        """
        Override this to add a Google CSX API Key
        """
        if self._google_csx_api_key is None:
            if hasattr(self.settings, 'clients') and hasattr(self.settings.clients, 'http_pool'):
                self._google_csx_api_key = getattr(self.settings.clients.http_pool, 'google_csx_api_key', None)
        return self._google_csx_api_key
    
    @property
    def google_csx_id(self) -> str:
        """
        Override this to add a Google CSX ID
        """
        if self._google_csx_id is None:
            if hasattr(self.settings, 'clients') and hasattr(self.settings.clients, 'http_pool'):
                self._google_csx_id = getattr(self.settings.clients.http_pool, 'google_csx_id', None)
        return self._google_csx_id
    

    @property
    def pdftotext_enabled(self) -> bool:
        """
        Returns whether pdftotext is enabled
        """
        if self._pdftotext_enabled is None:
            try:
                subprocess.check_output(['which', 'pdftotext'])
                self._pdftotext_enabled = True
            except Exception as e:
                self._pdftotext_enabled = False
        return self._pdftotext_enabled

    
    def cachify_get_name_builder_kwargs(self, func: str, **kwargs) -> Dict[str, Any]:
        """
        Gets the name builder kwargs
        """
        return {
            'include_http_methods': True,
            'special_names': ['pdftotext', 'csx'],
        }

    def __get_pdftotext(
        self,
        url: str,
        validate_url: Optional[bool] = False,
        raise_errors: Optional[bool] = None,
        **kwargs
    ) -> Optional[str]:
        """
        Transform a PDF File to Text directly from URL
        """
        if validate_url:
            validate_result = self._validate_url(url)
            if validate_result != True:
                if raise_errors: raise ValueError(f'Invalid URL: {url}. {validate_result}')
                self.logger.error(f'Invalid URL: {url}. {validate_result}')
                return None
        
        cmd = f'curl -s {url} | pdftotext -layout -nopgbrk -eol unix -colspacing 0.7 -y 58 -x 0 -H 741 -W 596 - -'
        process = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        try:
            stdout, stderr = process.communicate()
            stdout = stdout.decode('utf-8')
            return stdout
        except Exception as e:
            stderr = stderr.decode('utf-8')
            self.logger.error(f'Error in pdftotext: {stderr}: {e}')
            if raise_errors: raise e
            return None

    def _get_pdftotext(
        self,
        url: str,
        validate_url: Optional[bool] = False,
        retryable: Optional[bool] = False,
        retry_limit: Optional[int] = 3,
        raise_errors: Optional[bool] = None,
        **kwargs
    ) -> Optional[str]:
        """
        Transform a PDF File to Text directly from URL
        """
        if not self.pdftotext_enabled:
            raise ValueError('pdftotext is not enabled. Please install pdftotext')
        get_func = self.__get_pdftotext
        if retryable: get_func = http_retry_wrapper(max_tries = retry_limit + 1)(get_func)
        return get_func(url, raise_errors = raise_errors, **kwargs)


    async def __aget_pdftotext(
        self,
        url: str,
        validate_url: Optional[bool] = False,
        raise_errors: Optional[bool] = None,
        **kwargs
    ) -> Optional[str]:
        """
        Transform a PDF File to Text directly from URL
        """
        if validate_url:
            validate_result = await self._avalidate_url(url)
            if validate_result != True:
                if raise_errors: raise ValueError(f'Invalid URL: {url}. {validate_result}')
                self.logger.error(f'Invalid URL: {url}. {validate_result}')
                return None
        cmd = f'curl -s {url} | pdftotext -layout -nopgbrk -eol unix -colspacing 0.7 -y 58 -x 0 -H 741 -W 596 - -'
        process = await asyncio.subprocess.create_subprocess_shell(cmd, shell=True, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE)
        try:
            stdout, stderr = await process.communicate()
            stdout = stdout.decode('utf-8')
            return stdout
        except Exception as e:
            stderr = stderr.decode('utf-8')
            self.logger.error(f'Error in pdftotext: {stderr}: {e}')
            if raise_errors: raise e
            return None

    async def _aget_pdftotext(
        self,
        url: str,
        retryable: Optional[bool] = False,
        retry_limit: Optional[int] = 3,
        raise_errors: Optional[bool] = None,
        **kwargs
    ) -> Optional[str]:
        """
        Transform a PDF File to Text directly from URL
        """
        if not self.pdftotext_enabled:
            raise ValueError('pdftotext is not enabled. Please install pdftotext')
        get_func = self.__aget_pdftotext
        if retryable: get_func = http_retry_wrapper(max_tries = retry_limit + 1)(get_func)
        return await get_func(url, raise_errors = raise_errors, **kwargs)

    @cachify.register()
    def get_pdftotext(
        self,
        url: str,
        cachable: Optional[bool] = True,
        overwrite_cache: Optional[bool] = None,
        disable_cache: Optional[bool] = None,
        retryable: Optional[bool] = False,
        retry_limit: Optional[int] = 3,
        **kwargs
    ) -> Optional[str]:
        """
        Transform a PDF File to Text directly from URL
        """
        return self._get_pdftotext(url, retryable = retryable, retry_limit = retry_limit, **kwargs)

    @cachify.register()
    async def aget_pdftotext(
        self,
        url: str,
        cachable: Optional[bool] = True,
        overwrite_cache: Optional[bool] = None,
        disable_cache: Optional[bool] = None,
        retryable: Optional[bool] = False,
        retry_limit: Optional[int] = 3,
        **kwargs
    ) -> Optional[str]:
        """
        Transform a PDF File to Text directly from URL
        """
        return await self._aget_pdftotext(url, retryable = retryable, retry_limit = retry_limit, **kwargs)

    

    """
    Google CSX Methods
    """

    def get_csx(
        self,
        query: str,
        exact_terms: Optional[str] = None,
        exclude_terms: Optional[str] = None,
        file_type: Optional[str] = None,
        cachable: Optional[bool] = True,
        overwrite_cache: Optional[bool] = None,
        disable_cache: Optional[bool] = None,
        background: Optional[bool] = False,
        callback: Optional[Callable[..., Any]] = None,
        **kwargs,
    ) -> Dict[str, Union[List[Dict[str, Any]], Any]]:
        """
        Returns the Google CSX Results

        file_type: pdf
        """
        if not self.google_csx_api_key and not self.google_csx_id:
            raise ValueError('Google CSX API Key and ID are not set')
        params = {
            'key': self.google_csx_api_key, 'cx': self.google_csx_id, 'q': query,
        }
        if exact_terms: params['exactTerms'] = exact_terms
        if exclude_terms: params['excludeTerms'] = exclude_terms
        if file_type: 
            if 'application/' not in file_type: file_type = f'application/{file_type}'
            params['fileType'] = file_type
        if kwargs: params.update(kwargs)
        if background:
            return self.pooler.create_background(
                self.get,
                url = self.google_csx_base_url,
                return_type = 'json',
                params = params,
                task_callback = callback,
                cachable = cachable,
                overwrite_cache = overwrite_cache,
                disable_cache = disable_cache,
            )
        return self.get(url = self.google_csx_base_url, return_type = 'json', params = params, cachable = cachable, overwrite_cache = overwrite_cache, disable_cache = disable_cache)

    async def aget_csx(
        self,
        query: str,
        exact_terms: Optional[str] = None,
        exclude_terms: Optional[str] = None,
        file_type: Optional[str] = None,
        cachable: Optional[bool] = True,
        overwrite_cache: Optional[bool] = None,
        disable_cache: Optional[bool] = None,
        background: Optional[bool] = False,
        callback: Optional[Callable[..., Any]] = None,
        **kwargs,
    ) -> Dict[str, Union[List[Dict[str, Any]], Any]]:
        """
        Returns the Google CSX Results

        file_type: pdf
        """
        if not self.google_csx_api_key and not self.google_csx_id:
            raise ValueError('Google CSX API Key and ID are not set')
        params = {
            'key': self.google_csx_api_key, 'cx': self.google_csx_id, 'q': query,
        }
        if exact_terms: params['exactTerms'] = exact_terms
        if exclude_terms: params['excludeTerms'] = exclude_terms
        if file_type: 
            if 'application/' not in file_type: file_type = f'application/{file_type}'
            params['fileType'] = file_type
        if kwargs: params.update(kwargs)
        if background:
            return self.pooler.create_background(
                self.aget,
                url = self.google_csx_base_url,
                return_type = 'json',
                params = params,
                task_callback = callback,
                cachable = cachable,
                overwrite_cache = overwrite_cache,
                disable_cache = disable_cache,
            )
        # return await self._aget(url = self.google_csx_base_url, return_type = 'json', params = params)
        return await self.aget(url = self.google_csx_base_url, return_type = 'json', params = params, cachable = cachable, overwrite_cache = overwrite_cache, disable_cache = disable_cache)
    
        


from __future__ import annotations

"""
User Session
"""

import time

from pydantic import Field, model_validator
from lazyops.libs.abcs.utils.helpers import update_dict

from .base import BaseModel
from typing import Dict, Optional, List, Any, TYPE_CHECKING

class UserSession(BaseModel):
    """
    User Session Data
    """
    user_id: str
    api_key: str
    expiration_ts: Optional[int] = None
    data: Optional[Dict[str, Any]] = Field(default_factory = dict)

    model_config = {'arbitrary_types_allowed': True, 'extra': 'allow'}

    @model_validator(mode = 'after')
    def set_expiration_ts(self):
        """
        Sets the Expiration Timestamp
        """
        if not self.expiration_ts: self.expiration_ts = int(time.time()) + self.settings.user_session_expiration
        return self

    @property
    def is_expired(self) -> bool:
        """
        Returns True if the User Data is Expired
        """
        return self.expiration_ts < int(time.time())
    
    @property
    def ttl(self) -> int:
        """
        Returns the TTL
        """
        return max(self.expiration_ts - int(time.time()), 0)

    def update(self, **kwargs):
        """
        Updates the User Session Data
        """
        self.data = update_dict(self.data, kwargs)
        return self.data
    
    def get(self, key: str, default: Any = None) -> Any:
        """
        Gets the Item
        """
        return self.data.get(key, default)
    
    def pop(self, key: str, default: Any = None) -> Any:
        """
        Pops the Item
        """
        return self.data.pop(key, default)

    def __setitem__(self, key: str, value: Any):
        """
        Sets the Item
        """
        self.data[key] = value

    def __getitem__(self, key: str) -> Any:
        """
        Gets the Item
        """
        if self.data is None:
            self.data = {}
        return self.data[key]
    
    def __delitem__(self, key: str):
        """
        Deletes the Item
        """
        del self.data[key]

    def __contains__(self, key: str) -> bool:
        """
        Returns True if the Item is in the User Session
        """
        return key in self.data
    



from __future__ import annotations

"""
Utilities
"""
import functools
from lazyops.libs.pooler import ThreadPooler
from lazyops.libs.logging import logger
from typing import Any, Dict, Optional, TypeVar, Callable, TYPE_CHECKING

if TYPE_CHECKING:
    from .client import PostHogClient
    from .config import PostHogSettings


_ph_settings: Optional['PostHogSettings'] = None
_ph_client: Optional['PostHogClient'] = None

RT = TypeVar('RT')

def get_posthog_settings() -> 'PostHogSettings':
    """
    Returns the PostHog Settings
    """
    global _ph_settings
    if _ph_settings is None:
        from .config import PostHogSettings
        _ph_settings = PostHogSettings()
    return _ph_settings


def get_posthog_client(**kwargs) -> 'PostHogClient':
    """
    Returns the PostHog Client
    """
    global _ph_client
    if _ph_client is None:
        from .client import PostHogClient
        _ph_client = PostHogClient(**kwargs)
    elif kwargs: _ph_client.configure(**kwargs)
    return _ph_client

def register_posthog_client(client: 'PostHogClient', **kwargs):
    """
    Registers the PostHog Client
    """
    global _ph_client
    if _ph_client is None: _ph_client = client


def has_existing_posthog_client() -> bool:
    """
    Checks if there is an existing PostHog Client
    """
    global _ph_client
    return _ph_client is not None

"""
Basic Idea

@capture(event = 'test_event')
async def my_test_event(*args, ph_ctx: Dict[str, Any], **kwargs):

    # This will be passed to the function itself, which can be mutated
    # and will be sent to the Client after the function is called

    # ph_ctx = {'event': 'test_event', 'distinct_id': None, 'properties': {}}
    
    # Mutate the ph_ctx
    # ph_ctx['properties']['test'] = 'test'
    # ph_ctx['distinct_id'] = 'xxx'
    
    return 'test'

"""


def capture(
    event: Optional[str] = None,
    distinct_id: Optional[str] = None,
    properties: Optional[Dict[str, Any]] = None,
    **_kwargs: Any, 
) -> Callable[..., RT]:
    """
    Creates a decorator that can be used to capture events
    """
    def decorator(func: Callable[..., RT]) -> Callable[..., RT]:
        nonlocal event, distinct_id, properties

        client = _ph_client
        if client is None:
            ph_settings = get_posthog_settings()
            if ph_settings.is_enabled: client = get_posthog_client()
        if client is not None: 
            logger.info('Using Client for Capture')
            return client.capture(event = event, distinct_id = distinct_id, properties = properties, **_kwargs)(func)
        
        # Here we implement a deferred decorator
        if event is None: event = func.__name__
        properties = properties or {}
        ph_ctx = {'event': event, 'distinct_id': distinct_id, 'properties': properties, **_kwargs}
        logger.info('Using Deferred Capture')

        def _get_ph():
            """
            Gets the PostHog Client
            """
            nonlocal client
            if client is None: client = get_posthog_client()
            return client

        if ThreadPooler.is_coro(func):
            @functools.wraps(func)
            async def capture_decorator(*args, **kwargs):
                result = await func(*args, ph_ctx = ph_ctx, **kwargs)
                _get_ph().add_event(**ph_ctx)
                return result
        else:
            @functools.wraps(func)
            def capture_decorator(*args, **kwargs):
                result = func(*args, ph_ctx = ph_ctx, **kwargs)
                _get_ph().add_event(**ph_ctx)
                return result
        return capture_decorator
    return decorator
        


from __future__ import annotations

"""
Type Checking Support 
"""
import abc
import kr8s
from kr8s import objects
from kr8s.asyncio import objects as aobjects
from typing import Optional, Dict, Any, List, Union, overload, TYPE_CHECKING

if TYPE_CHECKING:
    ObjectT = Union[
        objects.Pod,
        objects.Service,
        objects.Deployment,
        objects.StatefulSet,
        objects.ConfigMap,
        objects.Secret,
        objects.PersistentVolume,
        objects.PersistentVolumeClaim,
        objects.Namespace,
        objects.Node,
        objects.Event,
        objects.Endpoints,
        objects.Ingress,
        objects.ServiceAccount,
        objects.Role,
        objects.RoleBinding,
        objects.ClusterRole,
        objects.APIObject,
    ]
    ObjectListT = List[ObjectT]
    aObjectT = Union[
        aobjects.Pod,
        aobjects.Service,
        aobjects.Deployment,
        aobjects.StatefulSet,
        aobjects.ConfigMap,
        aobjects.Secret,
        aobjects.PersistentVolume,
        aobjects.PersistentVolumeClaim,
        aobjects.Namespace,
        aobjects.Node,
        aobjects.Event,
        aobjects.Endpoints,
        aobjects.Ingress,
        aobjects.ServiceAccount,
        aobjects.Role,
        aobjects.RoleBinding,
        aobjects.ClusterRole,
        aobjects.APIObject,
    ]
    aObjectListT = List[aObjectT]




from __future__ import annotations

"""
Server CLI ABC
"""

import os
import sys
import typer
from typing import Optional, List, Dict, Any, Union, Type, TYPE_CHECKING

if TYPE_CHECKING:
    from lazyops.libs.abcs.configs.base import AppSettings


cmd = typer.Typer(no_args_is_help = True)

def build_server_cmd(
    cli_module: str,
    settings: 'AppSettings',
    host: Optional[str] = None,
    port: Optional[int] = None,
    num_server_workers: Optional[int] = 1,
    server_backend: Optional[str] = 'uvicorn',
    server_class: Optional[str] = 'uvloop',
    reload_enabled: Optional[bool] = False,
) -> str:
    """
    Prepare the Server Command
    """
    if not host: host = "0.0.0.0"
    if not port: port = 8080
    server_cmd = f"{server_backend} "

    # settings.server.app_host = host
    os.environ['APP_HOST'] = host
    # settings.server.app_port = port
    os.environ['APP_PORT'] = str(port)

    if server_backend == "uvicorn":
        server_cmd += f"{cli_module} --host {host} --port {port} --workers {num_server_workers} --h11-max-incomplete-event-size 524288"
        if reload_enabled: server_cmd += f" --reload --reload-dir {settings.module_path.as_posix()}"
        if server_class and server_class in {"auto", "asyncio", "uvloop"}: server_cmd += f" --loop {server_class}"

    elif server_backend == "hypercorn":
        server_cmd += f"{cli_module} --bind {host}:{port} --workers {num_server_workers}"
        if reload_enabled: server_cmd += " --reload"
        if server_class and server_class in {"asyncio", "uvloop", "trio"}: server_cmd += f" --worker-class {server_class}"

    elif server_backend == "gunicorn":
        server_cmd += f"--bind {host}:{port} --threads {num_server_workers}" #  --logger-class 'scout.utils.logs.StubbedGunicornLogger'"
        if reload_enabled: server_cmd += " --reload"
        if server_class:
            if server_class == "uvicorn":
                server_cmd += " --worker-class uvicorn.workers.UvicornWorker"
            elif server_class in {"sync", "eventlet", "gevent", "tornado", "gthread", "gaiohttp"}: server_cmd += f" --worker-class {server_class}"

        server_cmd += f" --pid {settings.global_ctx.state.server_process_id_path.as_posix()}"
        server_cmd += f" {cli_module}"

    else:
        typer.echo(f"Backend Server {server_backend} not supported")
        sys.exit(1)
    
    return server_cmd


"""
Tokens
"""
import time
from pydantic import BaseModel, model_validator
from typing import Optional, Union, List, Dict


class Token(BaseModel):
    access_token: str
    token_type: str
    expires_in: int
    expiration_ts: Optional[int] = None
    issued_at: Optional[int] = None

    @model_validator(mode = 'after')
    def validate_expiration(self):
        """
        Validates the expiration
        """
        if self.issued_at is None:
            self.issued_at = int(time.time())
        if self.expiration_ts is None:
            # self.expiration_ts = self.issued_at + 30
            self.expiration_ts = self.issued_at + (self.expires_in - 30)
        return self
        

    @property
    def is_expired(self) -> bool:
        """
        Returns True if the Token is Expired
        """
        return self.expiration_ts < int(time.time())
    
class AccessToken(Token):
    scope: str


class TokenPayload(BaseModel):
    sub: Optional[int] = None




"""
AuthZero FastAPI Integration
"""

from .utils.lazy import (
    get_az_settings,
    get_az_flow_schema,
    get_az_resource_schema,
    get_az_resource,
    get_az_mtg_api,
)

from .configs import AuthZeroSettings, settings as az_settings
from .types.auth import AuthZeroTokenAuth
from .types.user_roles import UserRole
from .types.security import (
    Authorization,
    APIKey,
)
from .types import errors
from .clients import (
    AuthZeroOAuthClient,
    AuthZeroAPIClient,
    CurrentUser,
    OptionalUser, 
    ValidUser, 
    get_current_user, 
    require_auth_role,
    require_roles,
    require_api_key,
    require_api_key_or_user_role,

    auth_role_dependency,
    api_key_or_user_role_dependency,
)


from __future__ import annotations

"""
Auth Zero User Dependencies
"""

from fastapi import Depends, Request
from fastapi.background import BackgroundTasks
from lazyops.libs.fastapi_utils.utils import create_function_wrapper
from ..types import errors
from ..types.current_user import CurrentUser, UserRole
from ..types.security import Authorization, APIKey
from ..utils.lazy import logger

from typing import Optional, List, Annotated, Type, Union


def extract_current_user(
    *args,
    _is_optional: Optional[bool] = False,
    **kwargs,
) -> CurrentUser:
    """
    Extract The Current User from the Endpoint Function
    """
    user_kw = 'user' if 'user' in kwargs else 'current_user'
    current_user: CurrentUser = kwargs.get(user_kw, None)
    if current_user is None and not _is_optional:
        raise errors.NoUserException()
    return current_user


def require_auth_role(
    role: Union[UserRole, str, int],
    enabled: Optional[bool] = True,
    dry_run: Optional[bool] = False,
    is_optional: Optional[bool] = False,
    verbose: Optional[bool] = True,
):
    """
    Creates an auth role validator wrapper

    Usage:

    >>> @app.get('/admin')
    >>> @require_auth_role(UserRole.ADMIN)
    >>> async def admin(current_user: CurrentUser):
            return {'admin': True}

    """

    if isinstance(role, (str, int)): role = UserRole.parse_role(role)
    def validation_func(*args, **kwargs):
        """
        Validation Function
        """
        if not enabled: return
        current_user = extract_current_user(*args, _is_optional = is_optional, **kwargs)
        if current_user.role < role:
            if verbose: logger.info(f'User {current_user.user_id} has role {current_user.role} which is less than {role}')
            if dry_run: return
            raise errors.InvalidAuthRoleException(detail = f'User {current_user.user_id} has role {current_user.role} which is less than or equal to {role}')
        return
    return create_function_wrapper(validation_func)

def require_roles(
    roles: Union[str, List[str]],
    require_all: Optional[bool] = False,
    dry_run: Optional[bool] = False,
    verbose: Optional[bool] = True,
):
    """
    Creates a role validator wrapper
    """
    if not isinstance(roles, list): roles = [roles]
    def validation_func(*args, **kwargs):
        """
        Validation Function
        """        
        current_user = extract_current_user(*args, **kwargs)
        if not current_user.has_user_roles(roles, require_all = require_all):
            if verbose: logger.info(f'User {current_user.user_id} does not have required roles: {roles} / {current_user.role}')
            if dry_run:  return
            raise errors.InvalidRolesException(detail = f'User {current_user.user_id} does not have required roles: {roles}')

    return create_function_wrapper(validation_func)


def get_current_user(
    required: Optional[bool] = True,
    roles_enabled: Optional[bool] = False,
    required_roles: Optional[List[str]] = None,
    require_all_roles: Optional[bool] = False,
    # api_keys: Optional[List[str]] = None,
    user_class: Optional[Type[CurrentUser]] = None,
    **kwargs,
):
    """
    Returns the current user
    """
    user_class = user_class or CurrentUser

    async def inner(
        request: Request, 
        background_tasks: BackgroundTasks, 
        api_key: APIKey = None, 
        authorization: Authorization = None
    ) -> Optional[CurrentUser]:
        """
        The inner method
        """
        current_user = user_class(
            required = required,
            **kwargs
        )
        await current_user(request, background_tasks, api_key = api_key, authorization = authorization)
        if current_user.is_valid:
            if roles_enabled and required_roles:
                if not current_user.has_user_roles(required_roles, require_all = require_all_roles):
                    logger.warning(f'User {current_user.user_id} does not have required roles: {required_roles}')
                    raise errors.InvalidRolesException(required_roles, require_all = require_all_roles)
            return current_user
        return None
    return inner

ValidUserDependency = Depends(get_current_user())
OptionalUserDependency = Depends(get_current_user(required = False))

OptionalUser = Annotated[Optional[CurrentUser], Depends(get_current_user(required = False))]
ValidUser = Annotated[CurrentUser, Depends(get_current_user())]


"""
These are able to be used as injections

@app.get('/')
@require_auth_role('admin')
async def index(current_user: ValidUser):
    ...


"""
def require_api_key(
    api_keys: Union[str, List[str]],
    dry_run: Optional[bool] = False,
    verbose: Optional[bool] = False,
    allow_authorized_user: Optional[bool] = False,
):
    """
    Creates an api key validator wrapper
    """
    if not isinstance(api_keys, list): api_keys = [api_keys]
    def has_api_key(*args, api_key: APIKey, current_user: OptionalUser, **kwargs):
        """
        Checks if the api key is valid
        """
        if api_key not in api_keys: 
            if allow_authorized_user and current_user and current_user.is_valid:
                return
            if verbose: logger.info(f'`{api_key}` is not a valid api key')
            if dry_run: return
            raise errors.InvalidAPIKeyException(detail = f'`{api_key}` is not a valid api key')
    
    return create_function_wrapper(has_api_key)


def require_api_key_or_user_role(
    api_keys: Union[str, List[str]],
    role: Optional[Union[str, UserRole]] = None,
    dry_run: Optional[bool] = False,
    verbose: Optional[bool] = False,
):
    """
    Creates an api key validator wrapper
    """
    if not isinstance(api_keys, list): api_keys = [api_keys]
    user_role = UserRole.parse_role(role) if role else None

    async def has_api_key_or_role(*args, current_user: OptionalUser, api_key: Optional[APIKey] = None, **kwargs):
        """
        Checks if the api key is valid
        """
        if not api_key and not current_user:
            if verbose: logger.info('No api key or user found')
            raise errors.NoAPIKeyException()
        if api_key and api_key in api_keys: return
        if current_user and current_user.is_valid:
            if user_role and current_user.role < user_role:
                if verbose: logger.info(f'User {current_user.user_id} does not have required role: {user_role}')
                if dry_run: return
                raise errors.InvalidAuthRoleException(detail = f'User {current_user.user_id} does not have required role: {user_role}')
            return
        if verbose: logger.info(f'`{api_key}` is not a valid api key')
        if dry_run: return
        raise errors.InvalidAPIKeyException(detail = f'`{api_key}` is not a valid api key')
    
    return create_function_wrapper(has_api_key_or_role)


"""
These require to be set as dependencies
"""

def auth_role_dependency(
    role: Union[str, UserRole],
    disabled: Optional[bool] = None,
    dry_run: Optional[bool] = False,
    verbose: Optional[bool] = False,
):
    """
    Creates an auth role validator wrapper
    """
    user_role = UserRole.parse_role(role) if isinstance(role, str) else role

    async def has_auth_role(current_user: ValidUser):
        """
        Checks if the auth role is valid
        """
        if disabled: return
        if current_user.role < user_role:
            if verbose: logger.info(f'User {current_user.user_id} does not have required role: {user_role}')
            if dry_run: return
            raise errors.InvalidAuthRoleException(detail = f'User {current_user.user_id} does not have required role: {user_role}')

    return Depends(has_auth_role)



def api_key_or_user_role_dependency(
    api_keys: Union[str, List[str]],
    role: Optional[Union[str, UserRole]] = None,
    dry_run: Optional[bool] = False,
    verbose: Optional[bool] = False,
):
    """
    Creates an api key validator wrapper
    """
    if not isinstance(api_keys, list): api_keys = [api_keys]
    user_role = UserRole.parse_role(role) if role else None

    async def has_api_key_or_role(
        current_user: OptionalUser,
        api_key: Optional[APIKey] = None,
    ):
        """
        Checks if the api key is valid
        """
        if not api_key and not current_user:
            if verbose: logger.info('No api key or user found')
            raise errors.NoAPIKeyException()
        if api_key and api_key in api_keys: return
        if current_user and current_user.is_valid:
            if user_role and current_user.role < user_role:
                if verbose: logger.info(f'User {current_user.user_id} does not have required role: {user_role}')
                if dry_run: return
                raise errors.InvalidAuthRoleException(detail = f'User {current_user.user_id} does not have required role: {user_role}')
            return
        if verbose: logger.info(f'`{api_key}` is not a valid api key')
        if dry_run: return
        raise errors.InvalidAPIKeyException(detail = f'`{api_key}` is not a valid api key')

    return Depends(has_api_key_or_role)
    


from __future__ import annotations

"""
Token Decoding Utilities
"""

from pydantic import ValidationError
from lazyops.libs import lazyload
from lazyops.libs.pooler import ThreadPooler
from .lazy import get_az_settings, get_az_resource
from typing import Any, Optional, List

if lazyload.TYPE_CHECKING:
    from fastapi import Request
    from fastapi.datastructures import Headers
    from ..configs import AuthZeroSettings
    from ..types.claims import UserJWTClaims


def decode_token(
    request: Optional['Request'] = None, 
    token: Optional[str] = None, 
    settings: Optional['AuthZeroSettings'] = None
) -> 'UserJWTClaims':
    """
    Attempts to decode the token from the request
    and allow for multi-audiences
    """
    from jose import jwt, JWTError
    assert request or token, 'Either a request or token must be provided'
    settings = settings or get_az_settings()
    token = token or get_auth_token(request.headers)
    err, jwt_claim = None, None
    audiences = settings.get_jwt_audiences()
    for aud in audiences:
        try:
            jwt_claim = jwt.decode(
                token = token,
                key = settings.jwks,
                algorithms = ['RS256'],
                audience = aud,
                issuer = settings.issuer,
            )
            # logger.info(f'Validated JWT Claim: {jwt_claim} against audience: {aud}')
            break
        except (JWTError, ValidationError) as e:
            err = e
    if jwt_claim is None:
        from ..types.errors import InvalidTokenException
        raise InvalidTokenException(error = e) from err
    return get_az_resource('user_jwt_claims', **jwt_claim)
    # return UserJWTClaims(**jwt_claim)

def decode_token_v1(
    request: Optional['Request'] = None, 
    token: Optional[str] = None, 
    settings: Optional['AuthZeroSettings'] = None
) -> 'UserJWTClaims':
    """
    Attempts to decode the token from the request
    """
    from jose import jwt, JWTError
    try:
        assert request or token, 'Either a request or token must be provided'
        settings = settings or get_az_settings()
        token = token or get_auth_token(request.headers)
        claims = UserJWTClaims(
            **jwt.decode(
                token = token,
                key=settings.jwks,
                algorithms = ['RS256'],
                audience = settings.audience,
                issuer = settings.issuer,
            )
        )
    except (JWTError, ValidationError) as e:
        from ..types.errors import InvalidTokenException
        raise InvalidTokenException(error = e) from e
    return claims



def get_auth_token(headers: 'Headers') -> str:
    """
    Gets the Auth Token from the Headers
    """
    settings = get_az_settings()
    authorization_header_value = headers.get(settings.authorization_header, headers.get(settings.authorization_header.capitalize()))
    if authorization_header_value:
        scheme, _, param = authorization_header_value.partition(" ")
        if scheme.lower() == settings.authorization_scheme:
            return param
    from ..types.errors import NoTokenException
    raise NoTokenException(detail = f'No Valid Authorization found on `{settings.authorization_header}')


def get_auth_token_v1(headers: 'Headers'):
    """
    Gets the Auth Token from the Headers
    """
    authorization_header_value = headers.get("Authorization")
    if authorization_header_value:
        scheme, _, param = authorization_header_value.partition(" ")
        if scheme.lower() == "bearer":
            return param
    from ..types.errors import NoTokenException
    raise NoTokenException(detail = 'Invalid Authorization Header')



from __future__ import annotations

"""
Kubernetes Client based on kr8s
"""
import os
import abc
import kr8s
import httpx
import base64
import functools
import contextlib
from pathlib import Path
from lazyops.libs.proxyobj.wraps import proxied
from lazyops.libs.logging import logger
from typing import Optional, Dict, TYPE_CHECKING
from .context import KubernetesContext
    
@proxied
class Kubernetes(abc.ABC):
    """
    The Kubernetes Client
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize the Kubernetes Client

        - For use in Kubernetes, the client will automatically use the in-cluster configuration
        - For use outside of Kubernetes, the client will use the kubeconfigs directory
          env: KUBECONFIGS_DIR
        """
        from lazyops.utils.system import is_in_kubernetes
        self.in_k8s = is_in_kubernetes()
        self.kconfigs_path = None
        if not self.in_k8s:
            if kctx_dir := os.getenv('KUBECONFIGS_DIR', os.getenv('KUBECONFIG_DIR', os.path.expanduser('~/.kube'))):
                self.kconfigs_path = Path(kctx_dir)
        self._ctxs: Dict[str, KubernetesContext] = {}
        self._default: Optional[str] = None
        self._kwargs = kwargs

    
    def find_kubeconfig_ctx(self, name: str) -> Optional[Path]:
        """
        Find a kubeconfig context
        """
        if self.kconfigs_path:
            for ctx_fname in self.kconfigs_path.iterdir():
                if name in ctx_fname.name:
                    logger.info(f'Using kubeconfig: `{ctx_fname.name}`')
                    return ctx_fname
        return None

    def get_or_init(
        self,
        name: Optional[str] = None, 
        url: Optional[str] = None,
        kubeconfig: Optional[str] = None,
        service_account: Optional[str] = None,
        context: Optional[str] = None,
        namespace: Optional[str] = None,
        set_as_default: Optional[bool] = False,
        **kwargs
    ) -> KubernetesContext:
        """
        Get or initialize a context
        """
        name = name if name is not None else (self._default or 'default')
        if name not in self._ctxs:
            if name != 'default' and not kubeconfig:
                kubeconfig = self.find_kubeconfig_ctx(name)
            self._ctxs[name] = KubernetesContext(
                name = name,
                url = url,
                kubeconfig = kubeconfig,
                service_account = service_account,
                context = context,
                namespace = namespace,
                **kwargs
            )
            if self._default is None or set_as_default:
                self._default = name
        return self._ctxs[name]

    @property
    def ctx(self) -> KubernetesContext:
        """
        Returns the default context
        """
        if self._default is None:
            return self.get_or_init()
        return self._ctxs[self._default]
    
    async def ainit(self) -> None:
        """
        Initialize the client
        """
        if not self._ctxs:
            await self.ctx.ainit()
        else:
            for ctx in self._ctxs.values():
                await ctx.ainit()

    def __getitem__(self, name: str) -> KubernetesContext:
        """
        Get a context
        """
        return self._ctxs[name]
    




from __future__ import annotations

"""
User Session Flows
"""
import time
from ..types.properties import StatefulProperty
from ..types.user_session import UserSession
from ..utils.lazy import logger
from .admin import AZManagementClient
from typing import List, Optional, Any, Dict, TYPE_CHECKING

if TYPE_CHECKING:
    from fastapi import Request
    from kvdb import PersistentDict
    from ..types.current_user import CurrentUser

class UserSessionFlow(StatefulProperty[UserSession]):
    """
    User Session Flow
    """

    name: Optional[str] = "user_session"
    schema_type = UserSession

    def __init__(
        self, 
        user_id: str,
        is_cache_key: Optional[bool] = None,
    ):
        if is_cache_key: cache_key = user_id
        else: cache_key = self.create_hash_key(user_id) if self.settings.session_user_id_hashed else user_id
        super().__init__(cache_key = cache_key)
        self._pdict_s: Optional['PersistentDict'] = None
        if is_cache_key: user_id = self.pdict_s.get(user_id)
        self.user_id = user_id

    @property
    def pdict_s(self) -> 'PersistentDict':
        """
        Returns the Persistent Dict for Session Mapping
        """
        if self._pdict_s is None: self._pdict_s = self.pdict.get_child('mapping')
        return self._pdict_s

    @property
    def session(self) -> UserSession:
        """
        Returns the Session
        """
        return self.resource
    
    @property
    async def asession(self) -> UserSession:
        """
        Returns the Session
        """
        return await self.aresource
    
    @property
    def session_data(self) -> Dict[str, Any]:
        """
        Returns the Session User Data
        """
        return self.load_data()
    
    @property
    async def asession_data(self) -> Dict[str, Any]:
        """
        Returns the User Data
        """
        return await self.aload_data()

    def fetch(self) -> 'UserSession':
        """
        Returns the Value
        """
        data = self.pdict.get(self.cache_key)
        if data is not None and (not self.is_valid_type(data) or self.is_expired(data)):
            self.delete()
            data = None
        return data
    
    async def afetch(self) -> 'UserSession':
        """
        Returns the Value
        """
        data = await self.pdict.aget(self.cache_key)
        if data is not None and (not self.is_valid_type(data) or self.is_expired(data)):
            await self.adelete()
            data = None
        return data
    

    def is_expired(self, data: UserSession) -> bool:
        """
        Returns True if the Data is Expired
        """
        return data.is_expired

    def get_ttl(self, data: UserSession) -> int:
        """
        Returns the TTL
        """
        return data.ttl
    
    def load_data(self) -> Dict[str, Any]:
        """
        Loads the Client Data
        """
        return self.pdict.get(self.data_cache_key, {})

    async def aload_data(self) -> Dict[str, Any]:
        """
        Loads the Client Data
        """
        return await self.pdict.aget(self.data_cache_key, {})
    
    def save_data(self, data: Dict[str, Any], ex: Optional[int] = None):
        """
        Saves the Data
        """
        self.pdict.set(self.data_cache_key, data, ex = ex)


    async def asave_data(self, data: Dict[str, Any], ex: Optional[int] = None):
        """
        Saves the Data
        """
        await self.pdict.aset(self.data_cache_key, data, ex = ex)

    def save(self, data: UserSession, ex: Optional[int] = None) -> None:
        """
        Saves the Object
        """
        super().save(data, ex = ex or self.get_ttl(data))
        self.pdict_s.set(self.cache_key, self.user_id, ex = ex or self.get_ttl(data))

    async def asave(self, data: UserSession, ex: Optional[int] = None) -> None:
        """
        Saves the Object
        """
        await super().asave(data, ex = ex or self.get_ttl(data))
        await self.pdict_s.aset(self.cache_key, self.user_id, ex = ex or self.get_ttl(data))

    def delete(self) -> None:
        """
        Deletes the Object
        """
        super().delete()
        self.pdict_s.delete(self.cache_key)

    async def adelete(self) -> None:
        """
        Deletes the Object
        """
        await super().adelete()
        await self.pdict_s.adelete(self.cache_key)

    @classmethod
    def get(
        cls, 
        user: 'CurrentUser',
        request: 'Request',
        **kwargs
    ) -> Optional[UserSession]:
        """
        Get the value
        """
        if user.settings.session_cookie_key not in request.cookies:
            return None
        session_key = request.cookies[user.settings.session_cookie_key]
        return cls(session_key, is_cache_key = True).fetch()
    
    @classmethod
    async def aget(
        cls, 
        user: 'CurrentUser',
        request: 'Request',
        **kwargs
    ) -> Optional[UserSession]:
        """
        Get the value
        """
        if user.settings.session_cookie_key not in request.cookies:
            # logger.info(f'No Session Cookie Found: {user.settings.session_cookie_key}')
            return None
        session_key = request.cookies[user.settings.session_cookie_key]
        # logger.info(f'Retrieving Session: {session_key}')
        return await cls(session_key, is_cache_key = True).afetch()

    if TYPE_CHECKING:

        @property
        def resource(self) -> UserSession:
            """
            Returns the Object
            """
            ...

        @property
        async def aresource(self) -> UserSession:
            """
            Returns the Object Resource
            """
            ...
        


from __future__ import annotations

"""
Hash Utilities
"""

import uuid
import hashlib
import base64
# from pydantic import ValidationError
# from lazyops.libs import lazyload
from lazyops.libs.pooler import ThreadPooler
# from lazyops.utils.helpers import lazy_import
from lazyops.imports._pycryptodome import resolve_pycryptodome
from urllib.parse import urlparse, parse_qsl, urlencode, urlunparse
from typing import Any, Optional, List, Dict

# if lazyload.TYPE_CHECKING:
#     import xxhash
#     import jose
#     # from Cryptodome.Cipher import AES
#     from fastapi import Request
#     # from fastapi.datastructures import Headers
#     # from jose import jwt, JWTError
#     from ..configs import AuthZeroSettings

# else:
#     xxhash = lazyload.LazyLoad("xxhash")
#     # jose = lazyload.LazyLoad('jose')
#     # jwt = jose.jwt
#     # JWTError = jose.JWTError
#     jwt = lazy_import('jose.jwt', is_module=True, allow_module=True)
#     JWTError = lazy_import('jose.exceptions.JWTError')
    # AES = lazy_import('Cryptodome.Cipher.AES', is_module=True, allow_module=True)
    # AES = lazyload.LazyLoad("Cryptodome.Cipher.AES", package = 'pycryptodomex')

    # jwt = lazyload.LazyLoad('jose.jwt')
    # JWTError = lazyload.LazyLoad('jose.exceptions.JWTError')

def create_uuid() -> str:
    """
    Returns a UUID
    """
    return str(uuid.uuid4())

def get_hashed_key(key: Any) -> str:
    """
    Returns a Hashed Key
    """
    return hashlib.sha256(str(key).encode()).hexdigest()
    # return xxhash.xxh128_hexdigest(str(key).encode())

def resize_key(key: str, length: int = 16) -> str:
    """
    Resizes the Key
    """
    return key.rjust((len(key) // length + 1) * length)

def encrypt_key(key: str, secret_key: str, access_key: str) -> str:
    """
    Encrypts the Key
    """
    resolve_pycryptodome(True)
    from Cryptodome.Cipher import AES

    cipher = AES.new(secret_key.encode(), AES.MODE_CBC, access_key.encode())
    string = cipher.encrypt(resize_key(key).encode())
    string = "".join("{:02x}".format(c) for c in string)
    return string

def decrypt_key(key: str, secret_key: str, access_key: str) -> str:
    """
    Decrypts the Key
    """
    resolve_pycryptodome(True)
    from Cryptodome.Cipher import AES

    cipher = AES.new(secret_key.encode(), AES.MODE_CBC, access_key.encode())
    string = cipher.decrypt(bytes.fromhex(key)).decode()
    return string.lstrip()

async def aencrypt_key(key: str, secret_key: str, access_key: str) -> str:
    """
    Encrypts the Key
    """
    return await ThreadPooler.run_async(
        encrypt_key,
        key,
        secret_key,
        access_key,
    )

async def adecrypt_key(key: str, secret_key: str, access_key: str) -> str:
    """
    Decrypts the Key
    """
    return await ThreadPooler.run_async(
        decrypt_key,
        key,
        secret_key,
        access_key,
    )

def parse_scopes(
    scope: Optional[str] = None,
    scopes: Optional[List[str]] = None,
) -> Optional[List[str]]:
    """
    Parses the Scopes
    """
    if scopes is None: scopes = []
    if scope is not None: 
        if " " in scope: scopes.extend(scope.split(' '))
        else: scopes.append(scope)
    return scopes or None

def normalize_audience_name(name: str) -> str:
    """
    Normalizes the audience name to transform the url

    >>> normalize_audience_name('https://domain.us.auth0.com/userinfo')
    'domain-us-auth0-com-userinfo'
    """
    return name.replace('https://', '').replace('http://', '').replace('/', '-').replace('.', '-').lower()

def create_code_challenge(secret_key: str) -> str:
    """
    Creates a code challenge
    """
    return base64.urlsafe_b64encode(hashlib.sha256(secret_key.encode()).digest()).decode().rstrip('=')

def encode_params_to_url(params: Dict[str, Any], url: str) -> str:
    """
    Encodes the params to a url
    """
    url_parts = list(urlparse(url))
    query = dict(parse_qsl(url_parts[4]))
    query.update(params)
    url_parts[4] = urlencode(query)
    return urlunparse(url_parts)



from __future__ import annotations

import abc
import datetime
import contextlib
from pydantic import BaseModel, model_validator, Field, PrivateAttr, validator
from lazyops.utils.logs import logger
from .static import SqliteTemplates
from .registry import get_or_register_sqlite_schema, get_or_register_sqlite_connection, retrieve_sqlite_model_schema, get_sqlite_model_pkey, get_or_register_sqlite_tablename, SQLiteModelConfig, get_sqlite_model_config
from .utils import normalize_sql_text
from typing import Optional, List, Tuple, Dict, Union, TypeVar, Any, overload, TYPE_CHECKING

if TYPE_CHECKING:
    import sqlite3
    # import aiosqlite
    from lazyops.libs import aiosqlite

    from pydantic.main import IncEx

    SQLResultT = TypeVar('SQLResultT', bound = 'SQLiteModelMixin')


class SQLiteModelMixin(BaseModel, abc.ABC):
    """
    The SQLite Model Mixin

    This mixin is used to create a SQLite model that can be used with SQLite. 
    
    It uses certain `Field` parameters to define the schema of the model.
    - `sql_pkey`: The primary key field of the model
    - `exclude`: Will exclude the field from the model
    - `include_sql`: If any of the fields contain this, it will ONLY select the fields that contain `include_sql`
    - `exclude_sql`: Will exclude the field from the SQLite model
    - `sql_search_precision`: The search precision of the field
    """

    # sql_config: Optional[SQLiteModelConfig] = Field(None, exclude = True, exclude_sql = True)
    _sql_config: Optional[SQLiteModelConfig] = PrivateAttr(None)

    @property
    def sql_config(self) -> SQLiteModelConfig:
        """
        Returns the SQLite Model Config
        """
        if self._sql_config is None:
            self._sql_config = get_sqlite_model_config(self.__class__)
        return self._sql_config
    
    @property
    def sql_pkey_column(self) -> str:
        """
        Returns the SQLite Model PKey
        """
        return self.sql_config.sql_pkey

    @property
    def sql_pkey(self) -> str:
        """
        Returns the SQLite Model PKey
        """
        return getattr(self, self.sql_pkey_column)
    
    @property
    def sql_tablename(self) -> str:
        """
        Returns the SQLite Tablename
        """
        return self.sql_config.tablename
    
    @property
    def sql_autoset(self) -> bool:
        """
        Returns the SQLite Autoset
        """
        return self.sql_config.autoset
    
    @property
    def sql_fields(self) -> List[str]:
        """
        Returns the SQLite Fields
        """
        return self.sql_config.sql_fields
    
    @property
    def sql_schema(self) -> Dict[str, Union[str, List[str], Dict[str, str]]]:
        """
        Returns the SQLite Schemas
        """
        return self.sql_config.sql_schema
    

    @classmethod
    def _get_sql_field_schema(cls) -> Tuple[Dict[str, str], str, Dict[str, int]]:
        # sourcery skip: low-code-quality
        """
        Gets the sql field schema
        """
        results = {}
        if any(
            f.json_schema_extra and f.json_schema_extra.get('include_sql', False) for f in cls.model_fields.values()
        ):
            
            field_keys = [
                name for name, f in cls.model_fields.items()
                if name != 'tablename' and \
                    (
                        f.json_schema_extra and f.json_schema_extra.get('include_sql', False) and \
                        not f.json_schema_extra.get('exclude_sql', False)
                    )
            ]
            logger.info(f'Model {cls.__name__} has `include_sql` fields. Only selecting fields: {field_keys}', prefix = f'|g|DB: {cls.__name__}|e|', colored = True)
        else:
            field_keys = [
                name for name, f in cls.model_fields.items()
                if name != 'tablename' and not f.exclude and \
                    (not f.json_schema_extra or not f.json_schema_extra.get('exclude_sql', False))
            ]
        pkey = [
            name for name in field_keys if \
                cls.model_fields[name].json_schema_extra and \
                cls.model_fields[name].json_schema_extra.get('sql_pkey', False)
        ]
        pkey = pkey[0] if pkey else field_keys[0]
        for name, field in cls.model_fields.items():
            # if field.exclude: continue
            if name == 'tablename': continue
            if name not in field_keys: continue
            if field.annotation in [str, Optional[str]]:
                results[name] = 'TEXT'
            elif field.annotation in [int, Optional[int]]:
                results[name] = 'INTEGER'
            elif field.annotation in [float, Optional[float]]:
                results[name] = 'REAL'
            elif field.annotation in [bool, Optional[bool]]:
                results[name] = 'BOOLEAN'
            elif field.annotation in [
                datetime.datetime,
                Optional[datetime.datetime],
            ]:
                results[name] = 'DATETIME'
            else:
                logger.warning(f"Unsupported Field Type: {field.annotation}", prefix = '|g|DB|e|', colored = True)
                continue
            if name == pkey: results[name] += ' PRIMARY KEY'
            # if name == pkey: results[name] += ' UNIQUE'
        search_precisions: Dict[str, int] = {
            name: (
                cls.model_fields[name].json_schema_extra.get('sql_search_precision', 0)
                if cls.model_fields[name].json_schema_extra else 0
            ) for name in field_keys
        }
        return results, pkey, search_precisions

    @classmethod
    def _get_sql_index_items(cls, **kwargs) -> List[Tuple]:
        """
        Gets the index items
        """
        return []
    

    @classmethod
    def execute_sql_init(
        cls, 
        conn: 'sqlite3.Connection',
        tablename: Optional[str] = None,
        skip_index: Optional[bool] = None,
        auto_set: Optional[bool] = None,
    ):
        """
        Executes the sql init

        Parameters
        ----------
        conn : sqlite3.Connection
            The connection to the database
        tablename : Optional[str], optional
            The table name to use, by default None. This must be passed at least once during initialization.
        skip_index : Optional[bool], optional
            Whether to skip the index creation, by default None. This must be passed at least once during initialization.
        auto_set : Optional[bool], optional
            Whether to automatically update the table whenever model attributes are updated, by default None.
        """
        cur = conn.cursor()
        schemas = get_or_register_sqlite_schema(cls, tablename, auto_set, conn)
        tablename = tablename or schemas['tablename']
        script = SqliteTemplates['init'].render(**schemas)
        # logger.info(f"SQL Script: {script}", prefix = f'|g|DB: {tablename}|e|', colored = True) 
        try:
            cur.executescript(script)
            conn.commit()
        except Exception as e:
            logger.error(f'[{tablename}] Error in sql init: {script}: {e}')
            raise e
        if not skip_index and (index_items := cls._get_sql_index_items()):
            # logger.info(f'Loaded {len(index_items)} items: {index_items[0]}')
            index_script = SqliteTemplates['index'].render(**schemas)
            # logger.info(f"SQL Script: {index_script}, {index_items[0]}", prefix = f'|g|DB: {tablename}|e|', colored = True)
            try:
                cur.executemany(index_script, index_items)
                conn.commit()
            except Exception as e:
                logger.error(f'[{tablename}] Error in indexing: {index_script}: {index_items[0]}: {e}')
                raise e
        # logger.info(f"Completed SQL Init: {tablename}", prefix = f'|g|DB: {tablename}|e|', colored = True)

    @classmethod
    async def aexecute_sql_init(
        cls, 
        conn: 'aiosqlite.Connection',
        tablename: Optional[str] = None,
        skip_index: Optional[bool] = None,
        auto_set: Optional[bool] = None,
    ):
        """
        Executes the sql init

        Parameters
        ----------
        conn : aiosqlite.Connection
            The connection to the database
        tablename : Optional[str], optional
            The table name to use, by default None. This must be passed at least once during initialization.
        skip_index : Optional[bool], optional
            Whether to skip the index creation, by default None. This must be passed at least once during initialization.
        auto_set : Optional[bool], optional
            Whether to automatically update the table whenever model attributes are updated, by default None.
        """
        schemas = get_or_register_sqlite_schema(cls, tablename, auto_set, conn)
        cur = await conn.cursor()
        await cur.executescript(SqliteTemplates['init'].render(**schemas))
        # cur = await conn.executescript(SqliteTemplates['init'].render(**schemas))
        if not skip_index and (index_items := cls._get_sql_index_items()):
            # logger.info(f'Loaded {len(index_items)} items: {index_items[0]}')
            index_script = SqliteTemplates['index'].render(**schemas)
            # logger.info(f"SQL Script: {index_script}", prefix = f'|g|DB: {tablename}|e|', colored = True)
            await cur.executemany(index_script, index_items)
            await conn.commit()

    @classmethod
    def _create_sql_query_from_kwargs(
        cls,
        schemas: Dict[str, Union[str, List[str], Dict[str, str]]],
        query: Optional[str] = None,
        **kwargs,
    ) -> str:
        """
        Creates the sql query from the kwargs

        precision: The search mode (0: loose, 1: field, 2: precise)
        """
        loose_matches: List[str] = []
        field_matches: List[str] = []
        exact_matches: List[str] = []
        search_precisions = schemas['search_precisions']
        for k, v in kwargs.items():
            if v is None: continue
            if isinstance(v, str) and not v.strip(): continue
            if k not in schemas['sql_fields']: continue
            precision = search_precisions.get(k, k)
            if precision == 0: loose_matches.append(normalize_sql_text(v))
            elif precision == 1: field_matches.append(f'({k} : {normalize_sql_text(v)})')
            elif precision == 2: exact_matches.append(f'{k} = "{v}"')
        
        query = query or ''
        tablename = schemas['tablename']

        if not loose_matches and not field_matches and not exact_matches and query:
            query = f'{tablename}_fts MATCH "{query}"'
            return query
        
        if exact_matches:
            query += f'({" OR ".join(exact_matches)}) AND '
        if loose_matches or field_matches:
            query += f'{tablename}_fts MATCH "'
            if field_matches: query += f'{" OR ".join(field_matches)} AND '
            if loose_matches: query += f'{" OR ".join(loose_matches)}'
            query = query.rstrip(' AND ')
            query += '"'
        query = query.replace('AND OR ', 'AND ')
        query = query.rstrip(' AND ')
        return query

    def _get_export_sql_data(
        self,
        schemas: Dict[str, Union[str, List[str], Dict[str, str]]],
        data: Optional[Dict[str, Any]] = None,
        include: 'IncEx' = None,
        exclude: 'IncEx' = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        **kwargs,
    ) -> Tuple[Union[str, int, float, Any], ...]:
        """
        Gets the export sql data
        """
        if data is None:
            data = self.model_dump(
                mode = 'json',
                include = include, 
                exclude = exclude, 
                by_alias = by_alias,
                exclude_unset = exclude_unset,
                exclude_defaults = exclude_defaults,
                exclude_none = exclude_none,
            )
        return tuple(data.get(k, None) for k in schemas['sql_keys'])

    def _get_export_sql_insert_fields(
        self,
        schemas: Dict[str, Union[str, List[str], Dict[str, str]]],
        data: Optional[Dict[str, Any]] = None,
        include: 'IncEx' = None,
        exclude: 'IncEx' = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        **kwargs,
    ) -> List[str]:
        """
        Gets the export sql index
        """
        if data is None:
            data = self.model_dump(
                mode = 'json',
                include = include, 
                exclude = exclude, 
                by_alias = by_alias,
                exclude_unset = exclude_unset,
                exclude_defaults = exclude_defaults,
                exclude_none = exclude_none,
            )
        return [k for k in schemas['sql_keys'] if k in data]


    def _get_export_sql_data_and_index(
        self,
        schemas: Dict[str, Union[str, List[str], Dict[str, str]]],
        include: 'IncEx' = None,
        exclude: 'IncEx' = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        **kwargs,
    ) -> Tuple[Dict[str, Union[str, List[str]]], Tuple[str, ...]]:
        """
        Gets the export sql data
        """
        data = self.model_dump(
            mode = 'json',
            include = include, 
            exclude = exclude, 
            by_alias = by_alias,
            exclude_unset = exclude_unset,
            exclude_defaults = exclude_defaults,
            exclude_none = exclude_none,
        )
        sql_insert_fields = [k for k in schemas['sql_keys'] if k in data]
        # sql_insert_values = tuple([data.get(k, None) for k in schemas['sql_keys']] + [None])
        sql_insert_values = tuple(data.get(k, None) for k in schemas['sql_keys'])
        results = {
            'sql_insert_query': ', '.join(sql_insert_fields),
            'sql_insert_fields': sql_insert_fields,
        }
        return (results, sql_insert_values)

    """
    Save Methods
    """

    @overload
    def sql_set(
        self,
        conn: 'sqlite3.Connection',
        tablename: Optional[str] = None,
        include: 'IncEx' = None,
        exclude: 'IncEx' = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        **kwargs,
    ) -> None:
        """
        Saves the data to the database
        """
        ...
    

    def sql_set(
        self,
        conn: 'sqlite3.Connection',
        tablename: Optional[str] = None,
        **kwargs,
    ):
        """
        Saves the data to the database
        """
        insert_dict, insert_data = self._get_export_sql_data_and_index(self.sql_schema, **kwargs)
        cur = conn.cursor()
        insert_script = SqliteTemplates['insert'].render(**self.sql_schema, **insert_dict)
        try:
            cur.execute(insert_script, insert_data)
            conn.commit()
            return cur.lastrowid
        except Exception as e:
            logger.error(f'[{self.sql_tablename}] Error in sql save: {insert_script}: {insert_data}: {e}')
            raise e


    @overload
    async def asql_set(
        self,
        conn: 'aiosqlite.Connection',
        tablename: Optional[str] = None,
        include: 'IncEx' = None,
        exclude: 'IncEx' = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        **kwargs,
    ) -> None:
        """
        Saves the data to the database
        """
        ...

    async def asql_set(
        self,
        conn: 'aiosqlite.Connection',
        tablename: Optional[str] = None,
        **kwargs,
    ):
        """
        Saves the data to the database
        """
        insert_dict, insert_data = self._get_export_sql_data_and_index(self.sql_schema, **kwargs)
        insert_script = SqliteTemplates['insert'].render(**self.sql_schema, **insert_dict)
        cur = await conn.cursor()
        try:
            await cur.execute(insert_script, insert_data)
            await conn.commit()
            return cur.lastrowid
        except Exception as e:
            logger.error(f'[{self.sql_tablename}] Error in sql save: {insert_script}: {insert_data}: {e}')
            raise e


    @classmethod
    def sql_set_many(
        cls: type['SQLResultT'],
        conn: 'sqlite3.Connection',
        items: List['SQLResultT'],
        tablename: Optional[str] = None,
        include: 'IncEx' = None,
        exclude: 'IncEx' = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        **kwargs,
    ):
        """
        Inserts many items to the database
        """
        schemas = get_or_register_sqlite_schema(cls, tablename)
        sql_insert_fields = items[0]._get_export_sql_insert_fields(schemas, include = include, exclude = exclude, by_alias = by_alias, exclude_unset = exclude_unset, exclude_defaults = exclude_defaults, exclude_none = exclude_none)
        sql_insert_values = [
            item._get_export_sql_data(schemas, include = include, exclude = exclude, by_alias = by_alias, exclude_unset = exclude_unset, exclude_defaults = exclude_defaults, exclude_none = exclude_none)
            for item in items
        ]
        insert_script = SqliteTemplates['insert'].render(**schemas, sql_insert_fields = sql_insert_fields)
        cur = conn.cursor()
        cur.executemany(insert_script, sql_insert_values)
        conn.commit()
        return cur.lastrowid
    
    @classmethod
    async def asql_set_many(
        cls: type['SQLResultT'],
        conn: 'aiosqlite.Connection',
        items: List['SQLResultT'],
        tablename: Optional[str] = None,
        include: 'IncEx' = None,
        exclude: 'IncEx' = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        **kwargs,
    ):
        """
        Inserts many items to the database
        """
        schemas = get_or_register_sqlite_schema(cls, tablename)
        sql_insert_fields = items[0]._get_export_sql_insert_fields(schemas, include = include, exclude = exclude, by_alias = by_alias, exclude_unset = exclude_unset, exclude_defaults = exclude_defaults, exclude_none = exclude_none)
        sql_insert_values = [
            item._get_export_sql_data(schemas, include = include, exclude = exclude, by_alias = by_alias, exclude_unset = exclude_unset, exclude_defaults = exclude_defaults, exclude_none = exclude_none)
            for item in items
        ]
        insert_script = SqliteTemplates['insert'].render(**schemas, sql_insert_fields = sql_insert_fields)
        cur = await conn.cursor()
        try:
            await cur.executemany(insert_script, sql_insert_values)
            await conn.commit()
            return cur.lastrowid
        except Exception as e:
            logger.error(f'[{tablename}] Error in sql save: {insert_script}: {sql_insert_values[:2]}: {e}')
            raise e
        # cur = await conn.executemany(insert_script, sql_insert_values)
        # await conn.commit()
        # return cur.lastrowid

    """
    Refresh Methods
    """

    def _set_sql_update_status(self, enabled: Optional[bool] = True):
        """
        Sets the sql update status
        """
        if enabled: 
            if not self.__pydantic_private__: self.__pydantic_private__ = {}
            self.__pydantic_private__['__in_sqlupdate__'] = True
        elif self.__pydantic_private__: _ = self.__pydantic_private__.pop('__in_sqlupdate__', None)

    def _is_in_sql_update(self) -> bool:
        """
        Checks if the object is in the sql update
        """
        if not self.__pydantic_private__: return False
        return '__in_sqlupdate__' in self.__pydantic_private__
        # return self.__pydantic_extra__.get('__in_sqlupdate__', False)

    @contextlib.contextmanager
    def _sql_update_context(self):
        """
        Context Manager for the sql update
        """
        self._set_sql_update_status(enabled = True)
        try:
            yield
        finally:
            self._set_sql_update_status(enabled = False)



    def sql_refresh(
        self: 'SQLResultT',
        conn: 'sqlite3.Connection',
        tablename: Optional[str] = None,
        **kwargs,
    )  -> 'SQLResultT':
        """
        Refreshes the data from the database
        """
        schemas = get_or_register_sqlite_schema(self.__class__, tablename)
        refresh_script = SqliteTemplates['refresh'].render(**schemas)
        cur = conn.cursor()
        cur.execute(refresh_script, (getattr(self, schemas['sql_pkey']),))
        result = cur.fetchone()
        if result is None: return None
        data = dict(zip(schemas['sql_keys'], result))
        with self._sql_update_context():
            for key, value in data.items():
                if value is None: continue
                if key in schemas['sql_keys']:
                    setattr(self, key, value)
        return self
        
    async def asql_refresh(
        self: 'SQLResultT',
        conn: 'aiosqlite.Connection',
        tablename: Optional[str] = None,
        **kwargs,
    )  -> 'SQLResultT':
        """
        Refreshes the data from the database
        """
        schemas = get_or_register_sqlite_schema(self.__class__, tablename)
        refresh_script = SqliteTemplates['refresh'].render(**schemas)
        cur = await conn.cursor()
        await cur.execute(refresh_script, (self.sql_pkey,))
        # await cur.execute(refresh_script, (getattr(self, schemas['sql_pkey']),))
        # cur = await conn.execute(refresh_script, (getattr(self, schemas['sql_pkey']),))
        result = await cur.fetchone()
        if result is None: return None
        data = dict(zip(schemas['sql_keys'], result))
        with self._sql_update_context():
            for key, value in data.items():
                if value is None: continue
                if key in schemas['sql_keys']:
                    setattr(self, key, value)
        return self

    """
    Search Methods
    """

    @overload
    @classmethod
    def search_one(
        cls: type[SQLResultT],
        conn: 'sqlite3.Connection',
        query: Optional[str] = None,
        skip: Optional[int] = None,
        **kwargs,
    ) -> 'SQLResultT':
        """
        Executes the sql search for one item
        """
        ...


    @overload
    @classmethod
    def search_one(
        cls: type[SQLResultT],
        conn: 'sqlite3.Connection',
        query: Optional[str] = None,
        skip: Optional[int] = None,
        return_fields: Optional[List[str]] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Executes the sql search for one item
        """
        ...

    @overload
    @classmethod
    def search_one(
        cls: type[SQLResultT],
        conn: 'sqlite3.Connection',
        query: Optional[str] = None,
        skip: Optional[int] = None,
        return_id_only: Optional[bool] = True,
        **kwargs,
    ) -> str:
        """
        Executes the sql search for one item
        """
        ...


    @classmethod
    def search_one(
        cls: type[SQLResultT],
        conn: 'sqlite3.Connection',
        query: Optional[str] = None,
        skip: Optional[int] = None,
        return_id_only: Optional[bool] = None,
        return_fields: Optional[List[str]] = None,
        tablename: Optional[str] = None,
        verbose: Optional[bool] = None,
        **kwargs,
    ) -> Optional[Union['SQLResultT', str]]:
        """
        Executes the sql search for one item
        """

        cur = conn.cursor()
        sql_config = get_sqlite_model_config(cls, tablename)
        tablename = tablename or sql_config.tablename
        query = cls._create_sql_query_from_kwargs(sql_config.sql_schema, query, **kwargs)
        if not query.strip():
            if verbose: logger.warning(f'[{tablename}] No query provided. Returning null: {query}, {kwargs}')
            return None
        
        query_script = SqliteTemplates['search'].render(
            query = query, limit = 1, skip = skip, 
            return_id_only = return_id_only, 
            return_fields = return_fields,
            **sql_config.sql_schema
        )
        try:
            cur.execute(query_script)
        except Exception as e:
            logger.error(f'[{tablename}] Error Searching: {query_script} {e}')
            raise e
        result = cur.fetchone()
        if result is None: return None
        if return_fields: return dict(zip(return_fields, result))
        if return_id_only: return result[0]
        data = dict(zip(sql_config.sql_keys, result))
        return cls.model_validate(data)


    @overload
    @classmethod
    def search_many(
        cls: type[SQLResultT],
        conn: 'sqlite3.Connection',
        query: Optional[str] = None,
        limit: Optional[int] = None,
        skip: Optional[int] = None,
        **kwargs,
    ) -> List['SQLResultT']:
        """
        Executes the sql search for many items
        """
        ...


    @classmethod
    def search_many(
        cls: type[SQLResultT],
        conn: 'sqlite3.Connection',
        query: Optional[str] = None,
        limit: Optional[int] = None,
        skip: Optional[int] = None,
        return_id_only: Optional[bool] = None,
        return_fields: Optional[List[str]] = None,
        tablename: Optional[str] = None,
        verbose: Optional[bool] = None,
        **kwargs,
    ) -> List[Union['SQLResultT', str]]:
        """
        Executes the sql search for many items
        """
        sql_config = get_sqlite_model_config(cls, tablename)
        tablename = tablename or sql_config.tablename
        query = cls._create_sql_query_from_kwargs(sql_config.sql_schema, query, **kwargs)
        if not query.strip():
            if verbose: logger.warning(f'[{tablename}] No query provided. Returning empty list: {query}, {kwargs}')
            return []

        query_script = SqliteTemplates['search'].render(
            query = query, limit = limit, skip = skip,
            return_id_only = return_id_only, return_fields = return_fields, **sql_config.sql_schema
        )
        cur = conn.cursor()
        try:
            cur.execute(query_script)
            results: List[Union['SQLResultT', str]] = []
            for result in cur.fetchall():
                if result:
                    if return_fields: results.append(dict(zip(return_fields, result)))
                    elif return_id_only: results.append(result[0])
                    else: results.append(cls.model_validate(dict(zip(sql_config.sql_keys, result))))
                else: results.append(None)
            return results
        except Exception as e:
            logger.error(f'[{tablename}] Error Searching: {query_script} {e}')
            raise e


    @overload
    @classmethod
    async def asearch_one(
        cls: type[SQLResultT],
        conn: 'sqlite3.Connection',
        query: Optional[str] = None,
        skip: Optional[int] = None,
        **kwargs,
    ) -> 'SQLResultT':
        """
        Executes the sql search for one item
        """
        ...

    @classmethod
    async def asearch_one(
        cls: type[SQLResultT],
        conn: 'aiosqlite.Connection',
        # conn: 'SQLiteAIOConnection',
        query: Optional[str] = None,
        skip: Optional[int] = None,
        return_id_only: Optional[bool] = None,
        return_fields: Optional[List[str]] = None,
        tablename: Optional[str] = None,
        verbose: Optional[bool] = None,
        **kwargs,
    ) -> Optional[Union['SQLResultT', str]]:
        """
        Executes the sql search for one item
        """
        sql_config = get_sqlite_model_config(cls, tablename)
        tablename = tablename or sql_config.tablename
        query = cls._create_sql_query_from_kwargs(sql_config.sql_schema, query, **kwargs)
        if not query.strip():
            if verbose: logger.warning(f'[{tablename}] No query provided. Returning null: {query}, {kwargs}')
            return None
        query_script = SqliteTemplates['search'].render(
            query = query, limit = 1, skip = skip,
            return_id_only = return_id_only, return_fields = return_fields, **sql_config.sql_schema
        )
        cursor = await conn.cursor()
        try:
            await cursor.execute(query_script)
            result = await cursor.fetchone()
        except Exception as e:
            logger.error(f'[{tablename}] Error Searching: {query_script} {e}')
            raise e
        if result is None: return None
        if return_fields: return dict(zip(return_fields, result))
        if return_id_only: return result[0]
        data = dict(zip(sql_config.sql_keys, result))
        return cls.model_validate(data)
    

    @overload
    @classmethod
    async def asearch_many(
        cls: type[SQLResultT],
        conn: 'sqlite3.Connection',
        query: Optional[str] = None,
        limit: Optional[int] = None,
        skip: Optional[int] = None,
        **kwargs,
    ) -> List['SQLResultT']:
        """
        Executes the sql search for many items
        """
        ...


    @classmethod
    async def asearch_many(
        cls: type[SQLResultT],
        conn: 'aiosqlite.Connection',
        # conn: 'SQLiteAIOConnection',
        query: Optional[str] = None,
        limit: Optional[int] = None,
        skip: Optional[int] = None,
        return_id_only: Optional[bool] = None,
        return_fields: Optional[List[str]] = None,
        tablename: Optional[str] = None,
        verbose: Optional[bool] = None,
        **kwargs,
    ) -> List[Union['SQLResultT', str]]:
        """
        Executes the sql search for one item
        """
        sql_config = get_sqlite_model_config(cls, tablename)
        tablename = tablename or sql_config.tablename
        query = cls._create_sql_query_from_kwargs(sql_config.sql_schema, query, **kwargs)
        if not query.strip():
            if verbose: logger.warning(f'[{tablename}] No query provided. Returning empty list: {query}, {kwargs}')
            return []
        
        query_script = SqliteTemplates['search'].render(
            query = query, limit = limit, skip = skip,
            return_id_only = return_id_only, return_fields = return_fields, **sql_config.sql_schema
        )
        results: List[Union['SQLResultT', str]] = []
        try:
            cursor = await conn.cursor()
            await cursor.execute(query_script)
            for result in await cursor.fetchall():
                if result:
                    if return_fields: results.append(dict(zip(return_fields, result)))
                    elif return_id_only: results.append(result[0])
                    else: results.append(cls.model_validate(dict(zip(sql_config.sql_keys, result))))
                else: results.append(None)
            return results
        except Exception as e:
            logger.error(f'[{tablename}] Error Searching: {query_script} {e}')
            raise e

    


    def __setattr__(self, name: str, value: Any) -> None:
        """
        Sets the attribute
        """
        super().__setattr__(name, value)
        # Skip setting attributes that start with `sql_` or `_`
        if name.startswith('sql_') or name.startswith('_'): return
        if not self.sql_autoset: return
        if name not in self.sql_fields: return
        if self._is_in_sql_update(): return
        # conn = get_or_register_sqlite_connection(self.__class__)
        update_script = SqliteTemplates['update_field_attribute'].render(**self.sql_schema, name = name)
        try:
            cur = self.sql_config.conn.cursor()
            # cur.execute(update_script, (value, getattr(self, self.sql_pkey),))
            cur.execute(update_script, (value, self.sql_pkey,))
            self.sql_config.conn.commit()
        except Exception as e:
            logger.error(f'[{self.sql_tablename}] Error in auto setting: {name} {value}: {update_script}: {e}')
            raise e


        

    """
    Delete Methods
    """

    def sql_delete(
        self,
        conn: 'sqlite3.Connection',
        tablename: Optional[str] = None,
        **kwargs,
    ):
        """
        Deletes the object from the database
        """
        # schemas = get_or_register_sqlite_schema(self.__class__, tablename)
        delete_script = SqliteTemplates['delete'].render(**self.sql_schema)
        cur = conn.cursor()
        cur.execute(delete_script, (self.sql_pkey,))
        conn.commit()


        # schemas = get_or_register_sqlite_schema(self.__class__, tablename)
        # delete_script = SqliteTemplates['delete'].render(**schemas)
        # cur = conn.cursor()
        # cur.execute(delete_script, (getattr(self, schemas['sql_pkey']),))
        # conn.commit()

    async def asql_delete(
        self,
        conn: 'aiosqlite.Connection',
        tablename: Optional[str] = None,
        **kwargs,
    ):
        """
        Deletes the object from the database
        """
    
        delete_script = SqliteTemplates['delete'].render(**self.sql_schema)
        cur = await conn.cursor()
        await cur.execute(delete_script, (self.sql_pkey,))
        await conn.commit()        

        # schemas = get_or_register_sqlite_schema(self.__class__, tablename)
        # delete_script = SqliteTemplates['delete'].render(**schemas)
        # cur = await conn.execute(delete_script, (getattr(self, schemas['sql_pkey']),))
        # await conn.commit()


from __future__ import annotations

"""
User data model for Auth0
"""

import time
from .base import BaseModel
from pydantic import Field, model_validator
from typing import Dict, Optional, List, Any

class UserDataBase(BaseModel):
    """
    Base User Data
    """
    username: Optional[str] = None
    family_name: Optional[str] = None
    given_name: Optional[str] = None
    name: Optional[str] = None
    email: Optional[str] = None
    email_verified: Optional[bool] = None
    nickname: Optional[str] = None
    phone_number: Optional[str] = None
    phone_verified: Optional[bool] = None
    picture: Optional[str] = None
    user_metadata: Optional[Dict[Any, Any]] = None
    app_metadata: Optional[Dict[Any, Any]] = None

class AZUserData(UserDataBase):
    """
    User Data
    """
    user_id: str
    multifactor: Optional[List[str]] = None
    expiration_ts: Optional[int] = None

    @model_validator(mode = 'after')
    def set_expiration_ts(self):
        """
        Sets the Expiration Timestamp
        """
        if self.expiration_ts is None: 
            self.expiration_ts = int(time.time()) + (self.settings.user_data_expiration or 60)
        return self
    

    @property
    def is_expired(self) -> bool:
        """
        Returns True if the User Data is Expired
        """
        return self.expiration_ts < int(time.time())



